"""
Created on Tue Jan 14 14:12:58 2014

@author: dave
"""
import os
import re

import numpy as np
import pandas as pd

from wetb.prepost import (mplutils, misc)


class dummy(object):
    def __init__(self, name='dummy'):
        self.__name__ = name

regex_units = re.compile('(\\[.*?\\])')


def ReadFileHAWCStab2Header(fname):
    """
    Read a file with a weird HAWCStab2 header that starts with a # (or not),
    and includes the column number and units between square brackets. Can also
    strip away the first integer number on the header line (e.g. .opt files).
    """

    def get_lines():
        # get the line that contains the header/column names and the first
        # line that holds the data
        line_header = f.readline()
        seek_data = f.tell()
        line_data = f.readline()
        # the opt file has a number of data points as the first element
        nrl = line_header.split()[0]
        try:
            int(nrl)
            # replace with empty text to no destroy the nice line-up of
            # the header with the data columns, assumed later
            line_header = line_header.replace(nrl, ' '*len(nrl))
        except ValueError:
            pass
        # sometimes there are more header lines. The header is always on the
        # last of the lines marked with #
        while line_data[:2].strip() == '#':
            line_header = line_data
            seek_data = f.tell()
            line_data = f.readline()

        return line_header, line_data, seek_data

    def get_col_widths(line):
        # it is very annoying that various files can have various column widths
        # also, the first column is one character wider than the rest
        i0 = re.search(r'\S',line).start()
        i1_col1 = line[i0:].find(' ') + i0
        # number of columns can also be different (gradients or not, node displ)
        nr_cols = int(round(len(line)/i1_col1, 0))
        colwidths = [i1_col1+1] + [i1_col1]*(nr_cols-1)
        return colwidths

    def get_col_names(line, colwidths):
        # because sometimes there are no spaces between the header of each column
        # sanitize the headers
        ci = np.array([0] + colwidths).cumsum()
        # remember zero based indexing
        ci[1:] = ci[1:] - 1
        columns = []
        for i in range(len(ci)-1):
            # also lose the index in the header, if there is one
            try:
                int(line.split()[-1])
                colname = line[ci[i]:ci[i+1]][:-2].replace('#', '').strip()
            except ValueError:
                colname = line[ci[i]:ci[i+1]].replace('#', '').strip()
            columns.append(colname)
        return columns

    with open(fname) as f:
        line_header, line_data, seek_data = get_lines()
        colwidths = get_col_widths(line_data)
        columns = get_col_names(line_header, colwidths)
        units = regex_units.findall(''.join(columns))
        # strip units from the column name, units is dictionary
        columns = [k.split('[')[0].strip() for k in columns]
        units = {col:unit for col, unit in zip(columns, units)}

        # gradients have duplicate columns: set for with wake updated
        # and another with frozen wake assumption, append _fw to the columns
        # used to indicate frozen wake gradients
        # the doubles occur at the end and are the following:
        # 'dQ/dt', 'dQ/dV', 'dQ/dO', 'dT/dt', 'dT/dV', 'dT/dO'
        if 'dQ/dt' in columns:
            i1 = columns.index('dQ/dt')
            if i1 > -1:
                i2 = columns.index('dQ/dt', i1+1)
            if i2 > i1:
                for i in range(i2, len(columns)):
                    columns[i] = columns[i] + '_fw'

        # we are at the second data line, go back to the start of the data
        f.seek(seek_data)
        df = pd.read_fwf(f, widths=colwidths, comment='#', header=None,
                         names=columns)

    return df, units


class InductionResults(object):
    """Column width can vary between versions and with/withouth gradient in
    output. Use get_col_width() for automatic detection.
    """
    def __init__(self, colwidth=None):
        """with gradients currently ind has columns width of 28 instead of 14!
        """
        self.cw = colwidth

    def get_col_width(self, fname):
        # figure out column width
        with open(fname) as fid:
            # line1 contains the header
            line1 = fid.readline()
            # line2 contains the numerical data
            line2 = fid.readline()

        # it is very annoying that various files can have various column widths
        # also, the first column is one character wider than the rest
        i0 = re.search(r'\S',line2).start()
        self.i1_col1 = line2[i0:].find(' ') + i0
        # number of columns can also be different (gradients or not, node displ)
        nr_cols = int(round(len(line2)/self.i1_col1, 0))
        self.colwidths = [self.i1_col1+1] + [self.i1_col1]*(nr_cols-1)

        # because sometimes there are no spaces between the header of each column
        # sanitize the headers
        ci = np.array([0] + self.colwidths).cumsum()
        # remember zero based indexing
        ci[1:] = ci[1:] - 1
        self.columns = []
        for i in range(len(ci)-1):
            # also lose the index in the header
            colname = line1[ci[i]:ci[i+1]][:-2].replace('#', '').strip()
            self.columns.append(colname)

    def read(self, fname):
        if self.cw is None:
            self.get_col_width(fname)
        self.wsp = int(fname.split('_u')[-1][:-4]) / 1000.0
        # self.df_data = pd.read_fwf(fname, header=0, widths=self.colwidths)
        self.df_data = pd.read_fwf(fname, skiprows=0, widths=self.colwidths)
        # we can not rely on read_fwf for the column names since for some
        # columns there is no space between one column and the next one.
        self.df_data.columns = self.columns


class results(object):
    """
    Loading HAWCStab2 result files
    """

    def __init__(self):
        pass

    def load_pwr(self, fname):
        pwr = np.atleast_2d(np.loadtxt(fname))

        res = dummy()

        res.wind = pwr[:,0]
        res.power = pwr[:,1]
        res.thrust = pwr[:,2]
        res.cp = pwr[:,3]
        res.ct = pwr[:,4]
        res.pitch_deg = pwr[:,8]
        res.rpm = pwr[:,9]

        return res

    def load_pwr_df(self, fname):
        return ReadFileHAWCStab2Header(fname)

    def load_cmb(self, fname):
        # aero-(servo)-elastic results for HS2>=2.14 have real_eig as 3th set
        with open(fname) as f:
            header = f.readline().replace('\n', '')
            cmb = np.loadtxt(f)
        # first column header has the units of the type of opearing points
        cols = misc.remove_items(header.split(']')[1].split(' '), '')
        nrmodes = np.array([int(k) for k in cols]).max()
        # when there is only data for one operating condition we only have one
        # row and consequently only a 1D array
        if len(cmb.shape) == 1:
            cmb = cmb.reshape( (1, cmb.shape[0]) )
        wind = cmb[:,0]
        freq = cmb[:,1:nrmodes+1]
        damp = cmb[:,nrmodes+1:nrmodes*2+1]
        real_eig = None
        if cmb.shape[1] > nrmodes*2+1:
            real_eig = cmb[:,nrmodes*2+1:]

        return wind, freq, damp, real_eig

    def load_cmb_df(self, fname):
        # index name can be rotor speed or wind speed
        with open(fname) as f:
            header = f.readline()
        oper_name = header.split('1')[0].strip().replace('#', '').lower()
        oper_name = oper_name.replace(' ', '').replace('[', '_')[:-1]
        oper_name = oper_name.replace('/', '')

        speed, freq, damp, real_eig = self.load_cmb(fname)
        mods = freq.shape[1]
        ops = freq.shape[0]

        df = pd.DataFrame(columns=[oper_name, 'Fd_hz', 'damp_ratio', 'mode'])
        df['Fd_hz'] = freq.flatten()
        df['damp_ratio'] = damp.flatten()
        if real_eig is not None:
            df['real_eig'] = real_eig.flatten()
        # now each mode number is a row so that means that each operating
        # point is now repeated as many times as there are modes
        df[oper_name] = speed.repeat(mods)
        modes = np.arange(1, mods+1, 1)
        df['mode'] = modes.reshape((1,mods)).repeat(ops, axis=0).flatten()

        return df

    def load_frf(self, fname, nr_inputs=3):
        frf = np.loadtxt(fname)

        self.nr_outputs = ((frf.shape[1] - 1) / 2) / nr_inputs
        self.nr_inputs = nr_inputs

        return frf

    def load_ind(self, fname, colwidth=None):
        """for results withouth gradients, colwidth=14, otherwise 28. Set to
        None to derive automatically.
        """
        ind = InductionResults(colwidth=colwidth)
        ind.read(fname)
        return ind.df_data

    def load_amp(self, fname):

        with open(fname) as f:
            line = f.readline()
            f.readline()
            line3 = f.readline()
            line4 = f.readline() # first data entry

        # assuming first we have: "# Mode number:"
        mode_nrs = line.split()[3:]

        elements_line1 = line4.split()
#        nrcols = len(elements_line1)
        element1 = elements_line1[0]
        width_col1 = line4.find(element1) + len(element1)

#        width = 14
#        nrcols = int((len(line)-1)/width)
#        # first columns has one extra character
#        # col nr1: rotor speed, col nr2: radius
#        widths = [width+1] + [width]*(nrcols-1)

        # the rotor/wind speed column does not have a unit
        ilocs = [0, width_col1-1] + [m.start() for m in re.finditer('\]', line3)]
        widths = np.diff(np.array(ilocs))
        # the length of the first element should be +1 due to zero based index
        widths[0] += 1

        # last line is empty
        df = pd.read_fwf(fname, header=2, widths=widths, skipfooter=1)
        units = regex_units.findall(''.join(df.columns))
        # no column number in the column name
        # since U_x, u_y, phase and theta will be repeated as many times as
        # there are modes, add the mode number in the column name
        columns = [k.replace('#', '').strip() for k in df.columns]
#        nrmodes = int((len(columns) - 2 )/6)
#        for k in range(nrmodes):
#            for i in range(6):
#                j = 2+k*6+i
#                columns[j] = columns[j].split('.')[0] + ' nr%i' % (k+1)

        for i, k in enumerate(mode_nrs):
            columns[i+2] = columns[i+2].split('.')[0] + ' nr%s' % k

        df.columns = columns

        return df, units

    def load_operation(self, fname):

        operation = np.loadtxt(fname, skiprows=1)
        # when the array is empty, set operation to an empty DataFrame
        if len(operation) == 0:
            cols = ['windspeed', 'pitch_deg', 'rotorspeed_rpm']
            return pd.DataFrame(columns=cols)
        # when there is only one data point, the array is 1D, we allways need
        # a 2D array otherwise the columns become rows in the DataFrame
        elif len(operation.shape) == 1:
            operation = operation.reshape((1, operation.shape[0]))
        try:
            cols = ['windspeed', 'pitch_deg', 'rotorspeed_rpm']
            operation = pd.DataFrame(operation, columns=cols)
        except ValueError:
            cols = ['windspeed', 'pitch_deg', 'rotorspeed_rpm', 'P_aero',
                    'T_aero']
            operation = pd.DataFrame(operation, columns=cols)
        return operation

    def load_matrices(self, fpath, basename, operating_point=1,
                      control_mat=False, local_wind_mat=False):
        """Load HAWCStab2 State Space system matrices

        The general file name format is:
        BASENAMETYPE_ase_ops_OPERATING_POINT_NUMBER.dat

        Where TYPE can be of the following:
            * amat, bmat, bvmat, cmat, dmat, dvmat, emat, fmat, fvmat

        Additionally, when including the control matrices:
            * BASENAMETYPE_ops_OPERATING_POINT_NUMBER.dat
            * TYPE: acmat, bcmat, ccmat, dcmat

        Or when including local wind speed
            * BASENAMETYPE_ase_ops_OPERATING_POINT_NUMBER.dat
            * TYPE: bvmat_loc_v, dvmat_loc_v, fvmat_loc_v


        Parameters
        ----------

        fpath : str

        basename : str

        operating_point : int, default=1


        Returns
        -------

        matrices : dict

        """
        mnames = ['amat', 'bmat', 'bvmat', 'cmat', 'dmat', 'dvmat', 'emat',
                  'fmat', 'fvmat']
        mnames_c = ['acmat', 'bcmat', 'ccmat', 'dcmat']
        mnames_v = ['bvmat_loc_v', 'dvmat_loc_v', 'fvmat_loc_v']

        if control_mat:
            mnames += mnames_c
        if local_wind_mat:
            mnames += mnames_v

        matrices = {}

        ase_template = '{:s}{:s}_ase_ops_{:d}.dat'
        ops_template = '{:s}{:s}_ops_{:d}.dat'

        for mname in mnames:
            rpl = (basename, mname, operating_point)
            template = ase_template
            if mname in mnames_c:
                template = ops_template
            fname = os.path.join(fpath, template.format(*rpl))
            matrices[mname] = np.loadtxt(fname)

        return matrices

    def load_cntrl_tuning(self, fname):
        """Load contruller tuning file.
        """

        tune = ReadControlTuning()
        tune.read_parameters(fname)
        tuning = tune.parameters2tags()
        return tuning

    def write_ae_sections_h2(self):
        """
        Get the aerosection positions from the HS2 ind result file and
        write them as outputs for HAWC2

        NOT IMPLEMENTED YET
        """
        self.ind


class ReadControlTuning(object):

    def __init__(self):
        """
        """
        self._aerogains = False

    def parse_line(self, line, controller):
        """Parses the output lines with the controller tuning parameters.
        Does not parse the aerodynamic gain lines.
        """

        if line.startswith('Aerodynamic gains'):
            self._aerogains = True
            return

        split1 = line.split('=')
        var1 = split1[0].strip()
        try:
            val1 = float(split1[1].split('[')[0])
            attr = getattr(self, controller)
            setattr(attr, var1, val1)

            if len(split1) > 2:
                var2 = split1[1].split(',')[1].strip()
                val2 = float(split1[2].split('[')[0])
                setattr(attr, var2, val2)
        except IndexError:
            pass

    def read_parameters(self, fpath):
        """
        Read the controller tuning file
        ===============================

        """

        with open(fpath, "r") as f:
            for i, line in enumerate(f):
                if i == 0:
                    controller = 'pi_gen_reg1'
                    setattr(self, controller, dummy())
                elif i == 2:
                    controller = 'pi_gen_reg2'
                    setattr(self, controller, dummy())
                elif i == 6:
                    controller = 'pi_pitch_reg3'
                    setattr(self, controller, dummy())
                elif i == 10:
                    controller = 'aero_damp'
                    setattr(self, controller, dummy())
                elif not self._aerogains:
                    self.parse_line(line, controller)
                # do not break since we want to see if aero gains are included
                # elif self._aerogains:
                #     break

        self.aero_gains_units = ['[deg]', '[kNm/deg]', '[kNm/deg]',
                                 '[kNm/(rad/s)]', '[kNm/(rad/s)]']
        self.aero_gains = pd.DataFrame()
        # in case the gains are missing from the file, don't try to read it
        if i > 17:
            arr = np.loadtxt(fpath, skiprows=17)
            columns = ['theta', 'dq/dtheta', 'dq/dtheta_fit', 'dq/domega',
                       'dq/domega_fit']
            # sometimes not all columns are present here, only take those that
            # are present
            self.aero_gains = pd.DataFrame(arr, columns=columns[:arr.shape[1]])

        # set some parameters to zero for the linear case, or when aerodynamic
        # gain scheduling is not used
        if not hasattr(self.pi_gen_reg2, 'Kd'):
            setattr(self.pi_gen_reg2, 'Kd', 0.0)
        if not hasattr(self.pi_pitch_reg3, 'Kd'):
            setattr(self.pi_pitch_reg3, 'Kd', 0.0)
        if not hasattr(self.pi_pitch_reg3, 'K2'):
            setattr(self.pi_pitch_reg3, 'K2', 0.0)
        if not hasattr(self.aero_damp, 'Kp2'):
            setattr(self.aero_damp, 'Kp2', 0.0)
        if not hasattr(self.aero_damp, 'Ko1'):
            setattr(self.aero_damp, 'Ko1', 0.0)
        if not hasattr(self.aero_damp, 'Ko2'):
            setattr(self.aero_damp, 'Ko2', 0.0)

    def parameters2tags(self):
        """Convert the tuning parameters into a dictionary whos keys are
        compatible with tag names in a HAWC2 master file.
        """

        tune_tags = {}

        tune_tags['[pi_gen_reg1.K]'] = self.pi_gen_reg1.K

        tune_tags['[pi_gen_reg2.I]'] = self.pi_gen_reg2.I
        tune_tags['[pi_gen_reg2.Kp]'] = self.pi_gen_reg2.Kp
        tune_tags['[pi_gen_reg2.Ki]'] = self.pi_gen_reg2.Ki
        tune_tags['[pi_gen_reg2.Kd]'] = self.pi_gen_reg2.Kd

        tune_tags['[pi_pitch_reg3.Kp]'] = self.pi_pitch_reg3.Kp
        tune_tags['[pi_pitch_reg3.Ki]'] = self.pi_pitch_reg3.Ki
        tune_tags['[pi_pitch_reg3.Kd]'] = self.pi_pitch_reg3.Kd
        tune_tags['[pi_pitch_reg3.K1]'] = self.pi_pitch_reg3.K1
        tune_tags['[pi_pitch_reg3.K2]'] = self.pi_pitch_reg3.K2

        tune_tags['[aero_damp.Kp2]'] = self.aero_damp.Kp2
        tune_tags['[aero_damp.Ko1]'] = self.aero_damp.Ko1
        tune_tags['[aero_damp.Ko2]'] = self.aero_damp.Ko2

        return tune_tags


def read_modid(fname):
    """Separate text file describing the modes.
    """

    df_modes = pd.read_csv(fname, comment='#', delimiter=';',
                           header=None, names=['mode_nr', 'description'],
                           converters={0:lambda x:x.strip(),
                                       1:lambda x:x.strip()})
    return df_modes['description'].values.tolist()


def read_cmb_all(f_cmb, f_pwr=None, f_modid=None, f_save=None, ps=[1,3,6]):
    """Convenience method to load Campbell and Performance data in one go

    Parameters
    ----------

    f_cmb : str

    f_pwr : str, default=None

    f_modid : str, default=None

    f_save : str, default=None

    ps : list, default=[1,2,3]


    Returns
    -------

    df_perf, df_freq, df_damp

    """

    hs2 = results()

    # create DataFrames with the freq/damp on the column name
    wind, freqs, damps, real_eig = hs2.load_cmb(f_cmb)
    nr_oper, nr_modes = freqs.shape

    # performance indicators, optional
    df_perf = None
    if f_pwr is not None:
        df_perf, units = hs2.load_pwr_df(f_pwr)
        nr_perf = df_perf.shape[0]
        if nr_oper != nr_perf or not np.allclose(wind, df_perf['V'].values):
            raise UserWarning('pwr and cmb files must have same operating points')

    # strip characters if there is a comment after the description
    if f_modid is None or not os.path.isfile(f_modid):
        modes_descr = ['{:02d}'.format(k+1) for k in range(nr_modes)]
    else:
        modes_descr = read_modid(f_modid)[:nr_modes]

    df_freq = pd.DataFrame(freqs, columns=modes_descr, index=wind)
    df_damp = pd.DataFrame(damps, columns=modes_descr, index=wind)

    if f_save is not None:
        tmp = pd.DataFrame(freqs, columns=modes_descr, index=wind)
        tmp.index.name = 'windspeed'
        # add p in frequencies
        if f_pwr is not None:
            for p in ps:
                tmp[f'{p}P'] = p*df_perf['Speed'].values / 60
        # sort columns on mean frequeny over wind speeds
        icolsort = tmp.values.mean(axis=0).argsort()
        tmp = tmp[tmp.columns[icolsort]]
        tmp.to_excel(f_save + '_freqs.xlsx')

        tmp = pd.DataFrame(damps, columns=modes_descr, index=wind)
        tmp.index.name = 'windspeed'
        tmp.to_excel(f_save + '_damps.xlsx')

    return df_perf, df_freq, df_damp


def plot_pwr(figname, fnames, labels=[], figsize=(11,7.15), dpi=120):

    hs2res = results()
    reslist = []
    if isinstance(fnames, list):
        if len(fnames) > 4:
            raise ValueError('compare up to maximum 4 HawcStab2 cases')
        for fname in fnames:
            reslist.append(hs2res.load_pwr(fname))
            # if the labels are not defined, take the file name
            if len(labels) < len(fnames):
                labels.append(os.path.basename(fname))
    else:
        reslist.append(hs2res.load_pwr(fname))

    colors = list('krbg')
    symbols = list('o<+x')
    alphas = [1.0, 0.9, 0.8, 0.75]

    fig, axes = mplutils.subplots(nrows=2, ncols=2, figsize=figsize,
                                   dpi=dpi, num=0)
    for i, res in enumerate(reslist):
        ax = axes[0,0]
        ax.plot(res.wind, res.power, color=colors[i],
                label='Power %s ' % labels[i],
                marker=symbols[i], ls='-', alpha=alphas[i])
        ax.set_title('Aerodynamic Power [kW]')#, RPM')

        ax = axes[0,1]
        ax.plot(res.wind, res.pitch_deg, color=colors[i],
                label='Pitch %s' % labels[i],
                marker=symbols[i], ls='-', alpha=alphas[i])
        ax.plot(res.wind, res.rpm, color=colors[i],
                label='RPM %s ' % labels[i],
                marker=symbols[i], ls='--', alpha=alphas[i])
        ax.set_title('Pitch [deg], RPM')

        ax = axes[1,0]
        ax.plot(res.wind, res.thrust, color=colors[i], label=labels[i],
                marker=symbols[i], ls='-', alpha=alphas[i])
        ax.set_title('Thrust [kN]')

        ax = axes[1,1]
        ax.plot(res.wind, res.cp, label='$C_p$ %s ' % labels[i], ls='-',
                color=colors[i], marker=symbols[i], alpha=alphas[i])
        ax.plot(res.wind, res.ct, label='$C_t$ %s ' % labels[i], ls='--',
                color=colors[i], marker=symbols[i], alpha=alphas[i])
        ax.set_title('Power and Thrust coefficients [-]')

    for ax in axes.ravel():
        ax.legend(loc='best')
        ax.grid(True)
        ax.set_xlim([res.wind[0], res.wind[-1]])
    fig.tight_layout()

    print('saving figure: %s ... ' % figname, end='')
    figpath = os.path.dirname(figname)
    if len(figpath)>0 and not os.path.exists(figpath):
        os.makedirs(figpath)
    fig.savefig(figname)
    fig.clear()
    print('done!')


class PlotCampbell(object):
    """
    Generic base class for a Campbell diagram plot. This class assumes a
    pandas DataFrame df_freq and df_damp is set by a subclass that deals
    with the tool dependen file IO. The following DataFrames are requried:
        * df_freq[operating points, modes] holds the respective mode frequencies
        * df_damp[operating points, modes] holds the respective mode damping
        * df_perf['wind'] holding the relevant wind speeds

    The column names are used as labels in the Campbell diagram.

    """

    alpha_box = 0.5

    def __init__(self, wind, df_freq, df_damp):
        self.wind = wind
        self.df_freq = df_freq
        self.df_damp = df_damp

    def _inplot_label_pos(self, nr_xpos, nr_series, xpos):
        """
        Generate sensible label positions
        """

        if xpos == 'random':
            pos = np.random.randint(1, nr_xpos-5, nr_series)
        elif xpos == 'centre':
            pos = np.zeros((nr_series,), dtype=int)
            pos[0:len(pos):2] = np.ceil(nr_xpos/4.0)
            pos[1:len(pos):2] = np.ceil(2.0*nr_xpos/4.0)
        elif xpos == 'borders':
            pos = np.zeros((nr_series,), dtype=int)
            pos[0:len(pos):2] = 2
            pos[2:len(pos):4] += 1
            pos[1:len(pos):2] = np.floor(3.3*nr_xpos/4.0)
            # and +1 alternating on the right
            pos[1:len(pos):4] += 1
            pos[3:len(pos):4] -= 1
        elif xpos == 'right':
            pos = np.zeros((nr_series,), dtype=int)
            pos[0:len(pos):2] = 2
            pos[1:len(pos):2] = np.ceil(1.0*nr_xpos/4.0)
        elif xpos == 'left':
            pos = np.zeros((nr_series,), dtype=int)
            pos[0:len(pos):2] = np.ceil(2.0*nr_xpos/4.0)
            pos[1:len(pos):2] = np.ceil(3.0*nr_xpos/4.0)
        elif xpos == 'spread':
            # TODO: we should determine the spacing based on wind speeds since
            # it might not be equally spaced
            pos = np.zeros((nr_series,), dtype=int)
            pos[0:len(pos):4] = 0
            pos[1:len(pos):4] = np.ceil(1.0*nr_xpos/4.0)
            pos[2:len(pos):4] = np.ceil(2.0*nr_xpos/4.0)
            pos[3:len(pos):4] = nr_xpos - 2

        return pos

    def plot_freq(self, ax, xpos='random', col='k', mark='^', ls='-',
                  modes='all'):
        """


        Parameters
        ----------
        ax : TYPE
            An Axes matplotlib instance that will be used for plotting.
        xpos : TYPE, str
            Specify strategy how to place the labels indifying the different modes.
            Valid string values are: 'random', 'centre', 'borders', 'right', 'left', 'spread'.
            The default is 'random'. If a list is passed, it should contain at
            least nr_modes elements. Items in the list refer to the index of the
            operating point (wind speed).
        col : string, optional
            Color used for the line plot. The default is 'k'. Can also be a list
            of colors in case modes should have different colors.
        mark : string, optional
            Marker to use for the line plot. The default is '^'. Can also be a list
            of mark symbols in case modes should have different symbols.
        ls : TYPE, optional
            Line style to use. The default is '-'. Can also be a list
            of line styles in case modes should have different line styles.
        modes : string, optional
            Identify which modes to plot. The default is 'all'. Optionally
            specify an integer to select the first x modes, or a list of indices
            to cherry pick modes of interest.

        Returns
        -------
        ax : matplotlib.axes
            An updated Axes matplotlib instance of the plot.

        """

        if isinstance(modes, str) and modes == 'all':
            df_freq = self.df_freq
        elif isinstance(modes, int):
            df_freq = self.df_freq[:modes]
        else:
            df_freq = self.df_freq[modes]

        nr_winds = df_freq.shape[0]
        nr_modes = df_freq.shape[1]

        pos = xpos
        if isinstance(xpos, str):
            pos = self._inplot_label_pos(nr_winds, nr_modes, xpos)
        elif isinstance(xpos, list) and len(xpos) < nr_modes:
            raise ValueError(f'xpos has only {len(xpos)}, while it needs to be >= {nr_modes}')

        if isinstance(col, str):
            col = [col]
        if isinstance(mark, str):
            mark = [mark]
        if isinstance(ls, str):
            ls = [ls]
        # just to make sure we always have enough colors/marks,lss
        col = col*nr_modes
        mark = mark*nr_modes
        ls = ls*nr_modes

        for i, (name, row) in enumerate(df_freq.items()):
            colmark = '%s%s%s' % (col[i], ls[i], mark[i])
            ax.plot(self.wind, row.values, colmark)#, mfc='w')
            x, y = self.wind[pos[i]], row.values[pos[i]]
            bbox = dict(boxstyle="round", alpha=self.alpha_box,
                        edgecolor=col[i], facecolor=col[i])
            ax.annotate(name, xy=(x, y), xycoords='data',
                        xytext=(-6, +20), textcoords='offset points',
                        fontsize=12, bbox=bbox, arrowprops=dict(arrowstyle="->",
                        connectionstyle="arc3,rad=.05"), color='w')

        self.nr_modes = nr_modes

        return ax

    def plot_damp(self, ax, xpos='random', col='r', mark='o' ,ls='--',
                  modes=14):
        """see plot_freq

        """

        # reduce the number of modes we are going to plot
        if isinstance(modes, str) and modes == 'all':
            df_damp = self.df_damp
        elif isinstance(modes, int):
            nr_modes = modes
            # sort the columns according to damping: lowest damped modes first
            # sort according to damping at lowest wind speed
            isort = self.df_damp.iloc[0,:].values.argsort()
            modes_sort_reduced = self.df_damp.columns[isort][:nr_modes]
            df_damp = self.df_damp[modes_sort_reduced]
        else:
            df_damp = self.df_damp[modes]

        nr_winds = df_damp.shape[0]
        nr_modes = df_damp.shape[1]

        if isinstance(col, str):
            col = [col]
        if isinstance(mark, str):
            mark = [mark]
        if isinstance(ls, str):
            ls = [ls]
        # just to make sure we always have enough colors/marks,lss
        col = col*nr_modes
        mark = mark*nr_modes
        ls = ls*nr_modes

        # put the labels in sensible places
        pos = xpos
        if isinstance(xpos, str):
            pos = self._inplot_label_pos(nr_winds, nr_modes, xpos)
        elif isinstance(xpos, list) and len(xpos) < nr_modes:
            raise ValueError(f'xpos has only {len(xpos)}, while it needs to be >= {nr_modes}')

        for i, (name, row) in enumerate(df_damp.items()):
            colmark = '%s%s%s' % (col[i], ls[i], mark[i])
            ax.plot(self.wind, row.values, colmark, alpha=0.8)#, mfc='w')
            x, y = self.wind[pos[i]], row.values[pos[i]]
            bbox = dict(boxstyle="round", alpha=self.alpha_box,
                        edgecolor=col[i], facecolor=col[i])
            ax.annotate(name, xy=(x, y), xycoords='data',
                        xytext=(-6, 20), textcoords='offset points',
                        fontsize=12, bbox=bbox, arrowprops=dict(arrowstyle="->",
                        connectionstyle="arc3,rad=.05"), color='w')

        self.nr_modes = nr_modes

        return ax

    def add_legend(self, ax, labels, on='freq'):
        if on == 'freq':
            i0s = self.istart_freqs
        else:
            i0s = self.istart_damps

        return ax.legend([ax.lines[k] for k in i0s[:-1]], labels, loc='best')


def plot_add_ps(ax, wind, rpm, col='g', fmax=10, ps=None):
    # plot 1P, 2P, ..., 9P
    bbox = dict(boxstyle="round", alpha=0.4, edgecolor=col, facecolor=col,)
    if ps is None:
        pmax = int(60*fmax/rpm.mean())
        ps = list(range(1, pmax))
    for p in ps:
        if p%3 == 0:
            alpha=0.6
            ax.plot(wind, rpm*p/60, '%s--' % col, )
        else:
            alpha = 0.4
            ax.plot(wind, rpm*p/60, '%s-.' % col)#, alpha=alpha)
        x, y = wind[10], rpm[10]*p/60
        p_str = '%iP' % p
        bbox['alpha'] = alpha
        ax.text(x, y, p_str, fontsize=9, verticalalignment='bottom',
                horizontalalignment='center', bbox=bbox, color='w')

    return ax


if __name__ == '__main__':

    pass



'''
Created on 20/01/2014

See documentation of HTCFile below

'''
from wetb.utils.process_exec import pexec
from wetb.hawc2.hawc2_pbs_file import HAWC2PBSFile
import jinja2
from wetb.utils.cluster_tools.os_path import fixcase, abspath, pjoin
from collections import OrderedDict
from wetb.hawc2.htc_contents import HTCContents, HTCSection, HTCLine
from wetb.hawc2.htc_extensions import HTCDefaults, HTCExtensions
import os


def fmt_path(path):
    return path.lower().replace("\\", "/")


class HTCFile(HTCContents, HTCDefaults, HTCExtensions):
    """Wrapper for HTC files

    Examples:
    ---------
    >>> htcfile = HTCFile('htc/test.htc')
    >>> htcfile.wind.wsp = 10
    >>> htcfile.save()

    #---------------------------------------------
    >>> htc = HTCFile(filename=None, modelpath=None) # create minimal htcfile

    #Add section
    >>> htc.add_section('hydro')

    #Add subsection
    >>> htc.hydro.add_section("hydro_element")

    #Set values
    >>> htc.hydro.hydro_element.wave_breaking = [2, 6.28, 1] # or
    >>> htc.hydro.hydro_element.wave_breaking = 2, 6.28, 1

    #Set comments
    >>> htc.hydro.hydro_element.wave_breaking.comments = "This is a comment"

    #Access section
    >>> hydro_element = htc.hydro.hydro_element #or
    >>> hydro_element = htc['hydro.hydro_element'] # or
    >>> hydro_element = htc['hydro/hydro_element'] # or
    >>> print (hydro_element.wave_breaking) #string represenation
    wave_breaking    2 6.28 1;    This is a comment
    >>> print (hydro_element.wave_breaking.name_) # command
    wave_breaking
    >>> print (hydro_element.wave_breaking.values) # values
    [2, 6.28, 1
    >>> print (hydro_element.wave_breaking.comments) # comments
    This is a comment
    >>> print (hydro_element.wave_breaking[0]) # first value
    2

    #Delete element
    htc.simulation.logfile.delete()
    #or
    del htc.simulation.logfile #Delete logfile line. Raise keyerror if not exists

    """

    filename = None
    jinja_tags = {}
    htc_inputfiles = []
    level = 0
    modelpath = "../"
    initial_comments = None

    def __init__(self, filename=None, modelpath=None, jinja_tags={}):
        """
        Parameters
        ---------
        filename : str
            Absolute filename of htc file
        modelpath : str
            Model path relative to htc file
        """
        if filename is not None:
            try:
                filename = fixcase(abspath(filename))
                with self.open(str(filename)):
                    pass
            except Exception:
                pass

            self.filename = filename

        self.jinja_tags = jinja_tags
        self.modelpath = modelpath or self.auto_detect_modelpath()

        if filename and self.modelpath != "unknown" and not os.path.isabs(self.modelpath):
            drive, p = os.path.splitdrive(os.path.join(os.path.dirname(str(self.filename)), self.modelpath))
            self.modelpath = os.path.join(drive, os.path.splitdrive(os.path.realpath(p))[1]).replace("\\", "/")
        if self.modelpath != 'unknown' and self.modelpath[-1] != '/':
            self.modelpath += "/"

        self.load()

    def auto_detect_modelpath(self):
        if self.filename is None:
            return "../"

        #print (["../"*i for i in range(3)])
        import numpy as np
        input_files = HTCFile(self.filename, 'unknown').input_files()
        if len(input_files) == 1:  # only input file is the htc file
            return "../"
        rel_input_files = [f for f in input_files if not os.path.isabs(f)]

        def isfile_case_insensitive(f):
            try:
                f = fixcase(f)  # raises exception if not existing
                return os.path.isfile(f)
            except IOError:
                return False
        found = ([np.sum([isfile_case_insensitive(os.path.join(os.path.dirname(self.filename), "../" * i, f))
                          for f in rel_input_files]) for i in range(4)])

        if max(found) > 0:
            relpath = "../" * np.argmax(found)
            return abspath(pjoin(os.path.dirname(self.filename), relpath))
        else:
            raise ValueError(
                "Modelpath cannot be autodetected for '%s'.\nInput files not found near htc file" % self.filename)

    def load(self):
        self.contents = OrderedDict()
        self.initial_comments = []
        self.htc_inputfiles = []
        if self.filename is None:
            lines = self.empty_htc.split("\n")
        else:
            lines = self.readlines(self.filename)

        lines = [l.strip() for l in lines]

        #lines = copy(self.lines)
        while lines:
            if lines[0].startswith(";"):
                self.initial_comments.append(lines.pop(0).strip() + "\n")
            elif lines[0].lower().startswith("begin"):
                self._add_contents(HTCSection.from_lines(lines))
            else:
                line = HTCLine.from_lines(lines)
                if line.name_ == "exit":
                    break
                self._add_contents(line)

    def readfilelines(self, filename):
        with self.open(self.unix_path(os.path.abspath(filename.replace('\\', '/'))), encoding='cp1252') as fid:
            txt = fid.read()
        if txt[:10].encode().startswith(b'\xc3\xaf\xc2\xbb\xc2\xbf'):
            txt = txt[3:]
        if self.jinja_tags:
            template = jinja2.Template(txt)
            txt = template.render(**self.jinja_tags)
        return txt.replace("\r", "").split("\n")

    def readlines(self, filename):
        if filename != self.filename:  # self.filename may be changed by set_name/save. Added it when needed instead
            self.htc_inputfiles.append(filename)
        htc_lines = []
        lines = self.readfilelines(filename)
        for l in lines:
            if l.lower().lstrip().startswith('continue_in_file'):
                filename = l.lstrip().split(";")[0][len("continue_in_file"):].strip().lower()

                if self.modelpath == 'unknown':
                    p = os.path.dirname(self.filename)

                    def isfile(f):
                        try:
                            return os.path.isfile(self.unix_path(f))
                        except OSError:
                            return False
                    lu = [isfile(os.path.abspath(os.path.join(p, "../" * i, filename)))
                          for i in range(4)].index(True)
                    filename = os.path.join(p, "../" * lu, filename)
                else:
                    filename = os.path.join(self.modelpath, filename)
                for line in self.readlines(filename):
                    if line.lstrip().lower().startswith('exit'):
                        break
                    htc_lines.append(line)
            else:
                htc_lines.append(l)
        return htc_lines

    def __setitem__(self, key, value):
        self.contents[key] = value

    def __str__(self):
        self.contents  # load
        return "".join(self.initial_comments + [c.__str__(1) for c in self] + ["exit;"])

    def save(self, filename=None):
        """Saves the htc object to an htc file.

        Args:
            filename (str, optional): Specifies the filename of the htc file to be saved.
            If the value is none, the filename attribute of the object will be used as the filename.
            Defaults to None.
        """
        self.contents  # load if not loaded
        if filename is None:
            filename = self.filename
        else:
            self.filename = filename
        # exist_ok does not exist in Python27
        if not os.path.exists(os.path.dirname(filename)) and os.path.dirname(filename) != "":
            os.makedirs(os.path.dirname(filename))  # , exist_ok=True)
        with self.open(filename, 'w', encoding='cp1252') as fid:
            fid.write(str(self))

    def set_name(self, name, subfolder=''):
        """Sets the base filename of the simulation files.

        Args:
            name (str): Specifies name of the log file, dat file (for animation), hdf5 file (for visualization) and htc file.
            subfolder (str, optional): Specifies the name of a subfolder to place the files in.
                If the value is an empty string, no subfolders will be created.
                Defaults to ''.

        Returns:
            None
        """
        # if os.path.isabs(folder) is False and os.path.relpath(folder).startswith("htc" + os.path.sep):
        self.contents  # load if not loaded

        def fmt_folder(folder, subfolder): return "./" + \
            os.path.relpath(os.path.join(folder, subfolder)).replace("\\", "/")

        self.filename = os.path.abspath(os.path.join(self.modelpath, fmt_folder(
            'htc', subfolder), "%s.htc" % name)).replace("\\", "/")
        if 'simulation' in self and 'logfile' in self.simulation:
            self.simulation.logfile = os.path.join(fmt_folder('log', subfolder), "%s.log" % name).replace("\\", "/")
            if 'animation' in self.simulation:
                self.simulation.animation = os.path.join(fmt_folder(
                    'animation', subfolder), "%s.dat" % name).replace("\\", "/")
            if 'visualization' in self.simulation:
                f = os.path.join(fmt_folder('visualization', subfolder), "%s.hdf5" % name).replace("\\", "/")
                self.simulation.visualization[0] = f
        elif 'test_structure' in self and 'logfile' in self.test_structure:  # hawc2aero
            self.test_structure.logfile = os.path.join(fmt_folder('log', subfolder), "%s.log" % name).replace("\\", "/")
        if 'output' in self:
            self.output.filename = os.path.join(fmt_folder('res', subfolder), "%s" % name).replace("\\", "/")

    def set_time(self, start=None, stop=None, step=None):
        self.contents  # load if not loaded
        if stop is not None:
            self.simulation.time_stop = stop
        else:
            stop = self.simulation.time_stop[0]
        if step is not None:
            self.simulation.newmark.deltat = step
        if start is not None:
            self.output.time = start, stop
            if "wind" in self:  # and self.wind.turb_format[0] > 0:
                self.wind.scale_time_start = start

    def expected_simulation_time(self):
        return 600

    def pbs_file(self, hawc2_path, hawc2_cmd, queue='workq', walltime=None,
                 input_files=None, output_files=None, copy_turb=(True, True)):
        walltime = walltime or self.expected_simulation_time() * 2
        if len(copy_turb) == 1:
            copy_turb_fwd, copy_turb_back = copy_turb, copy_turb
        else:
            copy_turb_fwd, copy_turb_back = copy_turb

        input_files = input_files or self.input_files()
        if copy_turb_fwd:
            input_files += [f for f in self.turbulence_files() if os.path.isfile(f)]

        output_files = output_files or self.output_files()
        if copy_turb_back:
            output_files += self.turbulence_files()

        return HAWC2PBSFile(hawc2_path, hawc2_cmd, self.filename, self.modelpath,
                            input_files, output_files,
                            queue, walltime)

    def input_files(self):
        self.contents  # load if not loaded
        if self.modelpath == "unknown":
            files = [str(f).replace("\\", "/") for f in [self.filename] + self.htc_inputfiles]
        else:
            files = [os.path.abspath(str(f)).replace("\\", "/") for f in [self.filename] + self.htc_inputfiles]
        if 'new_htc_structure' in self:
            for mb in [self.new_htc_structure[mb]
                       for mb in self.new_htc_structure.keys() if mb.startswith('main_body')]:
                if "timoschenko_input" in mb:
                    files.append(mb.timoschenko_input.filename[0])
                files.append(mb.get('external_bladedata_dll', [None, None, None])[2])
        if 'aero' in self:
            files.append(self.aero.ae_filename[0])
            files.append(self.aero.pc_filename[0])
            files.append(self.aero.get('external_bladedata_dll', [None, None, None])[2])
            files.append(self.aero.get('output_profile_coef_filename', [None])[0])
            if 'dynstall_ateflap' in self.aero:
                files.append(self.aero.dynstall_ateflap.get('flap', [None] * 3)[2])
            if 'bemwake_method' in self.aero:
                files.append(self.aero.bemwake_method.get('a-ct-filename', [None] * 3)[0])
        for dll in [self.dll[dll] for dll in self.get('dll', {}).keys() if 'filename' in self.dll[dll]]:
            files.append(dll.filename[0])
            f, ext = os.path.splitext(dll.filename[0])
            files.append(f + "_64" + ext)
            files.append(f + ".so")
        if 'wind' in self:
            files.append(self.wind.get('user_defined_shear', [None])[0])
            files.append(self.wind.get('user_defined_shear_turbulence', [None])[0])
            files.append(self.wind.get('met_mast_wind', [None])[0])
        if 'wakes' in self:
            files.append(self.wind.get('use_specific_deficit_file', [None])[0])
            files.append(self.wind.get('write_ct_cq_file', [None])[0])
            files.append(self.wind.get('write_final_deficits', [None])[0])
        if 'hydro' in self:
            if 'water_properties' in self.hydro:
                files.append(self.hydro.water_properties.get('water_kinematics_dll', [None])[0])
                files.append(self.hydro.water_properties.get('water_kinematics_dll', [None, None])[1])
        if 'soil' in self:
            if 'soil_element' in self.soil:
                files.append(self.soil.soil_element.get('datafile', [None])[0])
        try:
            dtu_we_controller = self.dll.get_subsection_by_name('dtu_we_controller')
            theta_min = dtu_we_controller.init.constant__5[1]
            if theta_min >= 90:
                files.append(os.path.join(os.path.dirname(
                    dtu_we_controller.filename[0]), "wpdata.%d" % theta_min).replace("\\", "/"))
        except Exception:
            pass

        try:
            files.append(self.force.dll.dll[0])
        except Exception:
            pass

        def fix_path_case(f):
            if os.path.isabs(f):
                return self.unix_path(f)
            elif self.modelpath != "unknown":
                try:
                    return "./" + os.path.relpath(self.unix_path(os.path.join(self.modelpath, f)),
                                                  self.modelpath).replace("\\", "/")
                except IOError:
                    return f
            else:
                return f
        return [fix_path_case(f) for f in set(files) if f]

    def output_files(self):
        self.contents  # load if not loaded
        files = []
        for k, index in [('simulation/logfile', 0),
                         ('simulation/animation', 0),
                         ('simulation/visualization', 0),
                         ('new_htc_structure/beam_output_file_name', 0),
                         ('new_htc_structure/body_output_file_name', 0),
                         ('new_htc_structure/struct_inertia_output_file_name', 0),
                         ('new_htc_structure/body_eigenanalysis_file_name', 0),
                         ('new_htc_structure/constraint_output_file_name', 0),
                         ('wind/turb_export/filename_u', 0),
                         ('wind/turb_export/filename_v', 0),
                         ('wind/turb_export/filename_w', 0)]:
            line = self.get(k)
            if line:
                files.append(line[index])
        if 'new_htc_structure' in self:
            if 'system_eigenanalysis' in self.new_htc_structure:
                f = self.new_htc_structure.system_eigenanalysis[0]
                files.append(f)
                files.append(os.path.join(os.path.dirname(f), 'mode*.dat').replace("\\", "/"))
            if 'structure_eigenanalysis_file_name' in self.new_htc_structure:
                f = self.new_htc_structure.structure_eigenanalysis_file_name[0]
                files.append(f)
                files.append(os.path.join(os.path.dirname(f), 'mode*.dat').replace("\\", "/"))
        files.extend(self.res_file_lst())

        for key in [k for k in self.contents.keys() if k.startswith("output_at_time")]:
            files.append(self[key]['filename'][0] + ".dat")
        return [f.lower() for f in files if f]

    def turbulence_files(self):
        self.contents  # load if not loaded
        if 'wind' not in self.contents.keys() or self.wind.turb_format[0] == 0:
            return []
        elif self.wind.turb_format[0] == 1:
            files = [self.get('wind.mann.filename_%s' % comp, [None])[0] for comp in ['u', 'v', 'w']]
        elif self.wind.turb_format[0] == 2:
            files = [self.get('wind.flex.filename_%s' % comp, [None])[0] for comp in ['u', 'v', 'w']]
        return [f for f in files if f]

    def res_file_lst(self):
        self.contents  # load if not loaded
        res = []
        for output in [self[k] for k in self.keys()
                       if self[k].name_.startswith("output") and not self[k].name_.startswith("output_at_time")]:
            dataformat = output.get('data_format', 'hawc_ascii')
            res_filename = output.filename[0]
            if dataformat[0] == "gtsdf" or dataformat[0] == "gtsdf64":
                res.append(res_filename + ".hdf5")
            elif dataformat[0] == "flex_int":
                res.extend([res_filename + ".int", os.path.join(os.path.dirname(res_filename), 'sensor')])
            else:
                res.extend([res_filename + ".sel", res_filename + ".dat"])
        return res

    def _simulate(self, exe, skip_if_up_to_date=False):
        self.contents  # load if not loaded
        if skip_if_up_to_date:
            from os.path import isfile, getmtime, isabs
            res_file = os.path.join(self.modelpath, self.res_file_lst()[0])
            htc_file = os.path.join(self.modelpath, self.filename)
            if isabs(exe):
                exe_file = exe
            else:
                exe_file = os.path.join(self.modelpath, exe)
            #print (from_unix(getmtime(res_file)), from_unix(getmtime(htc_file)))
            if (isfile(htc_file) and isfile(res_file) and isfile(exe_file) and
                str(HTCFile(htc_file)) == str(self) and
                    getmtime(res_file) > getmtime(htc_file) and getmtime(res_file) > getmtime(exe_file)):
                if "".join(self.readfilelines(htc_file)) == str(self):
                    return

        self.save()
        htcfile = os.path.relpath(self.filename, self.modelpath)
        assert any([os.path.isfile(os.path.join(f, exe)) for f in [''] + os.environ['PATH'].split(";")]), exe
        return pexec([exe, htcfile], self.modelpath)

    def simulate(self, exe, skip_if_up_to_date=False):
        errorcode, stdout, stderr, cmd = self._simulate(exe, skip_if_up_to_date)
        if ('simulation' in self.keys() and "logfile" in self.simulation and
                os.path.isfile(os.path.join(self.modelpath, self.simulation.logfile[0]))):
            with self.open(os.path.join(self.modelpath, self.simulation.logfile[0])) as fid:
                log = fid.read()
        else:
            log = "%s\n%s" % (str(stdout), str(stderr))

        if errorcode or 'Elapsed time' not in log:
            log_lines = log.split("\n")
            error_lines = [i for i, l in enumerate(log_lines) if 'error' in l.lower()]
            if error_lines:
                import numpy as np
                line_i = np.r_[np.array([error_lines + i for i in np.arange(-3, 4)]).flatten(),
                               np.arange(-5, 0) + len(log_lines)]
                line_i = sorted(np.unique(np.maximum(np.minimum(line_i, len(log_lines) - 1), 0)))

                lines = ["%04d %s" % (i, log_lines[i]) for i in line_i]
                for jump in np.where(np.diff(line_i) > 1)[0]:
                    lines.insert(jump, "...")

                error_log = "\n".join(lines)
            else:
                error_log = log
            raise Exception("\nError code: %s\nstdout:\n%s\n--------------\nstderr:\n%s\n--------------\nlog:\n%s\n--------------\ncmd:\n%s" %
                            (errorcode, str(stdout), str(stderr), error_log, cmd))
        return str(stdout) + str(stderr), log

    def simulate_hawc2stab2(self, exe):
        errorcode, stdout, stderr, cmd = self._simulate(exe, skip_if_up_to_date=False)

        if errorcode:
            raise Exception("\nstdout:\n%s\n--------------\nstderr:\n%s\n--------------\ncmd:\n%s" %
                            (str(stdout), str(stderr), cmd))
        return str(stdout) + str(stderr)

    def deltat(self):
        return self.simulation.newmark.deltat[0]

    def compare(self, other):
        if isinstance(other, str):
            other = HTCFile(other)
        return HTCContents.compare(self, other)

    @property
    def open(self):
        return open

    def unix_path(self, filename):
        filename = os.path.realpath(str(filename)).replace("\\", "/")
        ufn, rest = os.path.splitdrive(filename)
        ufn += "/"
        for f in rest[1:].split("/"):
            f_lst = [f_ for f_ in os.listdir(ufn) if f_.lower() == f.lower()]
            if len(f_lst) > 1:
                # use the case sensitive match
                f_lst = [f_ for f_ in f_lst if f_ == f]
            if len(f_lst) == 0:
                raise IOError("'%s' not found in '%s'" % (f, ufn))
            else:  # one match found
                ufn = os.path.join(ufn, f_lst[0])
        return ufn.replace("\\", "/")


#
#     def get_body(self, name):
#         lst = [b for b in self.new_htc_structure if b.name_=="main_body" and b.name[0]==name]
#         if len(lst)==1:
#             return lst[0]
#         else:
#             if len(lst)==0:
#                 raise ValueError("Body '%s' not found"%name)
#             else:
#                 raise NotImplementedError()
#

class H2aeroHTCFile(HTCFile):
    def __init__(self, filename=None, modelpath=None):
        HTCFile.__init__(self, filename=filename, modelpath=modelpath)

    @property
    def simulation(self):
        return self.test_structure

    def set_time(self, start=None, stop=None, step=None):
        if stop is not None:
            self.test_structure.time_stop = stop
        else:
            stop = self.simulation.time_stop[0]
        if step is not None:
            self.test_structure.deltat = step
        if start is not None:
            self.output.time = start, stop
            if "wind" in self and self.wind.turb_format[0] > 0:
                self.wind.scale_time_start = start


class SSH_HTCFile(HTCFile):
    def __init__(self, ssh, filename=None, modelpath=None):
        object.__setattr__(self, 'ssh', ssh)
        HTCFile.__init__(self, filename=filename, modelpath=modelpath)

    @property
    def open(self):
        return self.ssh.open

    def unix_path(self, filename):
        rel_filename = os.path.relpath(filename, self.modelpath).replace("\\", "/")
        _, out, _ = self.ssh.execute("find -ipath ./%s" % rel_filename, cwd=self.modelpath)
        out = out.strip()
        if out == "":
            raise IOError("'%s' not found in '%s'" % (rel_filename, self.modelpath))
        elif "\n" in out:
            raise IOError("Multiple '%s' found in '%s' (due to case senitivity)" % (rel_filename, self.modelpath))
        else:
            drive, path = os.path.splitdrive(os.path.join(self.modelpath, out))
            path = os.path.realpath(path).replace("\\", "/")
            return os.path.join(drive, os.path.splitdrive(path)[1])


if "__main__" == __name__:
    f = HTCFile(r"C:\mmpe\HAWC2\models\DTU10MWRef6.0\htc\DTU_10MW_RWT_power_curve.htc", "../")
    print(f.input_files())
#     f.save(r"C:\mmpe\HAWC2\models\DTU10MWRef6.0\htc\DTU_10MW_RWT_power_curve.htc")
#
#     f = HTCFile(r"C:\mmpe\HAWC2\models\DTU10MWRef6.0\htc\DTU_10MW_RWT.htc", "../")
#     f.set_time = 0, 1, .1
#     print(f.simulate(r"C:\mmpe\HAWC2\bin\HAWC2_12.8\hawc2mb.exe"))
#
#     f.save(r"C:\mmpe\HAWC2\models\DTU10MWRef6.0\htc\DTU_10MW_RWT.htc")



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 30 13:40:44 2020

@author: dave
"""

import os
from argparse import ArgumentParser
import math

import numpy as np
import pandas as pd

from wetb.dlc.high_level import Weibull_IEC


def aep_del(df_stats, resdir):
    """Works for regular wetb statistics df.
    """

    # note about the unique channel names: they are saved under
    # prepost-data/{sim_id}_unique-channel-names.csv
    # after running parpost.py

    # TODO: make option to have user define list of channels
    # channels = df_stats['channel'].unique()

    # DLC12: assumes that case_id (the file name) has the following format:
    # dlc10_wsp04_wdir000_s1001.htc

    # -----------------------------------------------------
    # USER INPUT
    neq_life = 1e7
    vref = 50 # used for the IEC Weibull distribution
    fact_1year = 365 * 24 * 0.975
    years_del = 20
    # specify which channel to use for the wind speed
    chwind = 'windspeed-global-Vy-0.00-0.00--119.00'
    # specify which channel to use for the electrical power output
    chpowe = 'DLL-generator_servo-inpvec-2'
    yawdist = {'wdir000':0.5, 'wdir010':0.25, 'wdir350':0.25}
    dlc10_dir = 'dlc10_powercurve'
    dlc12_dir = 'dlc12_iec61400-1ed3'
    # -----------------------------------------------------

    df_stats['hour_weight'] = 0

    # -----------------
    # DLC10
    # -----------------
    df_pc_dlc10 = pd.DataFrame()
    resdir_dlc10 = os.path.join(resdir, dlc10_dir)
    df_dlc10 = df_stats[df_stats['res_dir']==resdir_dlc10]
    if len(df_dlc10) > 0:
        df_pivot = df_dlc10.pivot(index='case_id', columns='channel', values='mean')
        df_pivot.sort_values(chwind, inplace=True)
        # wind speed and power
        wsp = df_pivot[chwind].values
        powe = df_pivot[chpowe].values
        # wind speed probability distribution for 1 year
        wsp_dist = Weibull_IEC(vref, wsp) * fact_1year
        # save data into the DataFrame table
        df_pc_dlc10['wsp'] = wsp
        df_pc_dlc10['powe'] = powe
        df_pc_dlc10['hours'] = wsp_dist
        df_pc_dlc10['aep'] = (powe * wsp_dist).sum()

    # -----------------
    # DLC12
    # -----------------
    resdir_dlc12 = os.path.join(resdir, dlc12_dir)
    df_dlc12 = df_stats[df_stats['res_dir']==resdir_dlc12].copy()
    df_pivot = df_dlc12.pivot(index='case_id', columns='channel', values='mean')
    # now add the wsp, yaw, seed columns again
    tmp = pd.DataFrame(df_pivot.index.values)
    df_pivot['dlc'] = ''
    df_pivot['wsp'] = ''
    df_pivot['yaw'] = ''
    df_pivot['seed'] = ''
    # assumes that case_id (the file name) has the following format:
    # dlc12_wsp04_wdir000_s1001.htc
    df_pivot[['dlc', 'wsp', 'yaw', 'seed']] = tmp[0].str.split('_', expand=True).values
    df_pivot['wsp'] = (df_pivot['wsp'].str[3:]).astype(int)

    # sorted on wsp, yaw, seed
    df_pivot.sort_index(inplace=True)

    # figure out how many seeds there are per yaw and wsp
    seeds = {'wsp_yaw':[], 'nr_seeds':[]}
    for ii, grii in df_pivot.groupby('wsp'):
        for jj, grjj in grii.groupby('yaw'):
            seeds['wsp_yaw'].append('%02i_%s' % (ii, jj))
            seeds['nr_seeds'].append(len(grjj))

    nr_seeds = np.array(seeds['nr_seeds']).max()
    seeds_min = np.array(seeds['nr_seeds']).min()
    assert nr_seeds == seeds_min

    # yaw error probabilities
    for yaw, dist in yawdist.items():
        df_pivot.loc[df_pivot['yaw']==yaw, 'hour_weight'] = dist/nr_seeds
    df_pivot['annual_hours_wsp'] = 0

    # Weibull hour distribution as function of wind speeds
    wsps = df_pivot['wsp'].unique()
    wsps.sort()
    wsp_dist = Weibull_IEC(vref, wsps) * fact_1year
    for i, wsp in enumerate(wsps):
        sel = df_pivot['wsp']==wsp
        df_pivot.loc[sel, 'annual_hours_wsp'] = wsp_dist[i]

    df_pivot['annual_hour_dist'] = df_pivot['annual_hours_wsp'] * df_pivot['hour_weight']

    # check we still have the same amount of hours in a year
    hours1 = df_pivot['annual_hour_dist'].sum()
    hours2 = wsp_dist.sum()
    assert np.allclose(hours1, hours2)

    aep = (df_pivot['annual_hour_dist'] * df_pivot[chpowe]).sum()

    df_pc_dlc12 = df_pivot[[chwind, chpowe, 'dlc', 'wsp', 'yaw', 'seed',
                            'hour_weight', 'annual_hours_wsp', 'annual_hour_dist']].copy()
    df_pc_dlc12['aep'] = aep

    # -----------------
    # DLC12 DEL
    ms = [k for k in df_dlc12.columns if k.find('m=') > -1]
    dict_Leq = {m:[] for m in ms}
    dict_Leq['channel'] = []

    # TODO: make option to have user define list of channels
    # for channel in channels:
        #statsel = df_dlc12[df_dlc12['channel']==channel].copy()

    for channel, grch in df_dlc12.groupby('channel'):
        # sort the case_id values in the same way as the pivot table is to
        # align the hours with the corresponding case
        statsel = grch.copy()
        statsel.set_index('case_id', inplace=True)
        statsel = statsel.loc[df_pivot.index,:]

        dict_Leq['channel'].append(channel)

        for m in ms:
            m_ = float(m.split('=')[1])
            eq1hz_mod = np.power(statsel[m].values, m_)
            # R_eq_mod will have to be scaled from its simulation length
            # to 1 hour (hour distribution is in hours...). Since the
            # simulation time has not been multiplied out of R_eq_mod yet,
            # we can just multiply with 3600 (instead of doing 3600/neq)
            tmp = (eq1hz_mod * df_pivot['annual_hour_dist'] * years_del * 3600).sum()
            # the effective Leq for each of the material constants
            leq = math.pow(tmp/neq_life, 1.0/m_)
            dict_Leq[m].append(leq)

    df_leq = pd.DataFrame(dict_Leq)
    df_leq.set_index('channel', inplace=True)

    return df_pc_dlc10, df_pc_dlc12, df_leq


if __name__ == '__main__':

    parser = ArgumentParser(description = "Calculate AEP and LEQ")
    parser.add_argument('--res', type=str, default='res', action='store',
                        dest='resdir', help='Directory containing result files')
    opt = parser.parse_args()

    sim_id = os.path.basename(os.getcwd())
    fname = f'prepost-data/{sim_id}_statistics.h5'
    df_stats = pd.read_hdf(fname, key='table')

    df_pc_dlc10, df_pc_dlc12, df_leq= aep_del(df_stats, opt.resdir)

    # save fot Excel files
    df_leq.to_excel(f'prepost-data/{sim_id}_del.xlsx')
    df_pc_dlc10.to_excel(f'prepost-data/{sim_id}_aep_10.xlsx')
    df_pc_dlc12.to_excel(f'prepost-data/{sim_id}_aep_12.xlsx')



'''
Created on 20. jul. 2017

@author: mmpe
'''
import unittest
from wetb.utils.test_files import get_test_file
from wetb.wind.turbulence import mann_turbulence
from wetb.wind.turbulence.mann_turbulence import parameters2name
from tests import npt


class TestMannTurbulence(unittest.TestCase):

    def testLoad(self):
        fn = get_test_file('h2a8192_8_8_16384_32_32_0.15_10_3.3u.dat')
        self.assertRaises(AssertionError, mann_turbulence.load, fn, (31, 31))
        self.assertRaises(AssertionError, mann_turbulence.load, fn, (8192, 32, 32))
        u = mann_turbulence.load(fn, (8192, 8, 8))
        self.assertEqual(u.shape, (8192, 8 * 8))

    def test_loaduvw(self):
        fn = 'h2a8192_8_8_16384_32_32_0.15_10_3.3%s.dat'
        fn_lst = [get_test_file(fn % uvw) for uvw in 'uvw']

        u, v, w = mann_turbulence.load_uvw(fn_lst, (8192, 8, 8))
        self.assertEqual(u.shape, (8192, 8 * 8))
        self.assertTrue(u.shape == v.shape == w.shape == (8192, 8 * 8))
        u, v, w = mann_turbulence.load_uvw(fn_lst[0].replace("u.dat", "%s.dat"), (8192, 8, 8))
        self.assertEqual(u.shape, (8192, 8 * 8))
        self.assertTrue(u.shape == v.shape == w.shape == (8192, 8 * 8))

    def test_parameters2name(self):
        self.assertEqual(parameters2name((8192, 8, 8), (16384, 32, 32), 0.15, 10, 3.3, 1, 1)[
                         0], './turb/mann_l10.0_ae0.1500_g3.3_h1_8192x8x8_2.000x4.00x4.00_s0001u.turb')

    def test_save(self):
        fn = get_test_file('h2a8192_8_8_16384_32_32_0.15_10_3.3u.dat')
        u_ref = mann_turbulence.load(fn, (8192, 8, 8))
        mann_turbulence.save(u_ref, 'tmp_u.turb')
        u = mann_turbulence.load('tmp_u.turb', (8192, 8, 8))
        npt.assert_array_equal(u, u_ref)


#     def test_name2parameters(self):
#         self.assertEqual(name2parameters('./turb/mann_l10.0_ae0.15_g3.3_h1_8192x8x8_2.000x4.00x4.00_s0001u.turb'),
#                          ((8192, 8, 8), (16384., 32., 32.), 0.15, 10, 3.3, 1, 1))


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



from tests import npt
import pytest
import numpy as np
from wetb.fast.fast_io import load_output, load_binary_output
import os

testfilepath = os.path.join(os.path.dirname(__file__), 'test_files/')  # test file path


def test_load_output():
    data, info = load_output(testfilepath + 'DTU10MW.out')
    npt.assert_equal(data[4, 3], 4.295E-04)
    npt.assert_equal(info['name'], "DTU10MW")
    npt.assert_equal(info['attribute_names'][1], "RotPwr")
    npt.assert_equal(info['attribute_units'][1], "kW")


def test_load_binary():
    data, info = load_output(testfilepath + 'test_binary.outb')
    npt.assert_equal(info['name'], 'test_binary')
    npt.assert_equal(info['description'], 'Modified by mwDeriveSensors on 27-Jul-2015 16:32:06')
    npt.assert_equal(info['attribute_names'][4], 'RotPwr')
    npt.assert_equal(info['attribute_units'][7], 'deg/s^2')
    npt.assert_almost_equal(data[10, 4], 138.822277739535)


def test_load_binary_buffered():
    # The old method was not using a buffer and was also memory expensive
    # Now use_buffer is set to true by default
    import numpy as np
    data, info = load_binary_output(testfilepath + 'test_binary.outb', use_buffer=True)
    data_old, info_old = load_binary_output(testfilepath + 'test_binary.outb', use_buffer=False)
    npt.assert_equal(info['name'], info_old['name'])
    np.testing.assert_array_equal(data[0, :], data_old[0, :])
    np.testing.assert_array_equal(data[-1, :], data_old[-1, :])


@pytest.mark.parametrize('fid,tol', [(2, 1), (3, 1e-4)])
@pytest.mark.parametrize('buffer', [True, False])
def test_load_bindary_fid(fid, tol, buffer):
    data2, info2 = load_output(testfilepath + '5MW_Land_BD_DLL_WTurb_fid04.outb')
    data, info = load_binary_output(testfilepath + '5MW_Land_BD_DLL_WTurb_fid%02d.outb' % fid, use_buffer=buffer)
    for k, v in info2.items():
        if k not in {'name', 'description'}:
            npt.assert_array_equal(info[k], v)
    r = data.max(0) - data.min(0) + 1e-20
    npt.assert_array_less(np.abs(data - data2).max(0), r * tol)


def test_load_output2():
    data, info = load_output(testfilepath + 'DTU10MW.out')
    npt.assert_equal(info['name'], "DTU10MW")
    npt.assert_equal(info['attribute_names'][1], "RotPwr")
    npt.assert_equal(info['attribute_units'][1], "kW")


def test_load_output3():
    # This file has an extra comment at the end
    data, info = load_output(testfilepath + 'FASTOut_Hydro.out')
    npt.assert_almost_equal(data[3, 1], -1.0E+01)



'''
Created on 12/09/2013

@author: mmpe
'''

import h5py
import numpy as np
from wetb import gtsdf

import unittest
import os
from wetb.gtsdf.gtsdf import collect_statistics
import pytest

tmp_path = os.path.dirname(__file__) + "/tmp/"
tfp = os.path.dirname(os.path.abspath(__file__)) + "/test_files/"


class Test_gsdf(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        if not os.path.isdir(tmp_path):
            os.makedirs(tmp_path)

    @classmethod
    def tearDownClass(cls):
        super(Test_gsdf, cls).tearDownClass()
        # shutil.rmtree(tmp_path)

    def test_gtsdf_stat(self):
        # test_gtsdf_stat
        time, data, info = gtsdf.load(tfp + 'test.hdf5')
        fn = tmp_path + "test_stat.hdf5"
        gtsdf.save(fn, data, time=time, **info)
        gtsdf.add_statistic(fn)
        stat_data, info = gtsdf.load_statistic(fn)
        self.assertEqual(data[:, 0].min(), stat_data.values[0, 0])
        self.assertEqual(stat_data.shape, (49, 10))

        # test_gtsdf_compress2stat
        time, data, info = gtsdf.load(tfp + 'test.hdf5')
        fn = tmp_path + "test_compress2stat.hdf5"
        gtsdf.save(fn, data, time=time, **info)
        del info['dtype']
        gtsdf.save(tmp_path + "test_compress2stat2.hdf5", data, time=time, dtype=float, **info)
        gtsdf.compress2statistics(fn)
        self.assertLess(os.path.getsize(fn) * 50, os.path.getsize(tfp + 'test.hdf5'))

        # test_collect_stat
        with pytest.raises(Exception, match=r'No \*\.hdf5 files found in'):
            collect_statistics('missing', tmp_path)

        df, info = collect_statistics('.', tmp_path, filename='*stat.hdf5')
        assert df.shape == (98, 12)
        df, info = collect_statistics('.', tmp_path + "..", filename='*stat.hdf5')
        assert df.shape == (98, 12)
        with pytest.raises(Exception, match=r'No \*stat\.hdf5 files found in'):
            collect_statistics('.', tmp_path + "..", filename='*stat.hdf5', recursive=False)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 25 16:15:25 2018

@author: shfe@dtu.dk

Static p-y soil model and cyclic degradation model
"""

import numpy as np

def py_static(y, pu, yc):
    """
    Static p-y model developed by Matlock
    
    Parameters
    ----------------
    y -- displacement
    pu -- ultimate stress
    yc -- ultimate displacement
    
    Output
    ----------------
    p -- soil resistence pressure
    """
    
    if y/yc <= 8:
        p = 0.5*pu*(y/yc)**(1/3)
        
    else:
        p = pu
        
    return p
    
def strain_static(p, pu, yc):
    """
    Static strain using p-y model developed by Matlock
    
    Parameters
    ----------------
    p -- stress
    pu -- ultimate stress
    yc -- ultimate displacement
    
    Output
    ----------------
    y -- soil strain
    """
    if p <= pu:
        y = (p/(0.5*pu))**3*yc
    
    else:
        raise ValueError("The applied stress is LARGER than ultimate strength")
    
    return y
    
def load_unload_py(p, pu, yc):
    """
    load/unload module using p-y model developed by Matlock
    
    Parameters
    ----------------
    p -- stress
    pu -- ultimate stress
    yc -- ultimate displacement
    
    Output
    ----------------
    y -- soil strain
    y_cum -- cumulative soil displacement
    """
    
    y = strain_static(p, pu, yc)
    
    if p <= 0.5*pu:  # elastic
        y_cum = 0
        
    else:
        y_cum = y - p/(0.5*pu/yc)
        
    return y, y_cum



    
def degra_coeff(N, y1, b):
    """
    Degradation factor for p-y model
    
    Parameters
    ----------------
    N -- number of cycle
    y1 -- displacement predicted by static p-y curve
    b -- pile diameter
    
    Output
    ----------------
    lambda -- Degradation factor
    """
    # check

    lambda_N = np.abs(y1/(0.2*b)*np.log(N))
    
    try:
        lambda_N <= 1
    except:
        print("soil must be degraded!!!")
    
    return lambda_N
    
def py_cyclic(y, pu, yc, N, y1, b):
    """
    Degradation p-y model with cyclic load
    
    Parameters
    ----------------
    y -- displacement
    pu -- ultimate stress
    yc -- ultimate displacement
    N -- number of cycle
    y1 -- displacement predicted by static p-y curve
    b -- pile diameter
    
    Output
    ----------------
    p -- degraded soil resistence pressure
    """
    
    lambda_N = degra_coeff(N, y1, b)
    p = py_static(y, pu, yc) * (1 - lambda_N)

    return p

def py_parcel(load_parcel, cycle_parcel, pu, yc, b):
    """
    Degradation p-y model after a parcel of cyclic load
    
    Parameters
    ----------------
    load_parcel -- load parcel
    cycle_parcel -- number of cycle of load parcel
    yc -- ultimate displacement
    b -- pile diameter
    
    Output
    ----------------
    p -- degraded soil resistence pressure
    """
    
    pu_N = np.zeros(np.size(load_parcel)+1)
    pu_N[0] = pu
    lambda_N = np.zeros(np.size(load_parcel))
    y_N = np.zeros(np.size(load_parcel))
    
    for i in np.arange(np.size(load_parcel)):
        load_i = load_parcel[i]
        cycle_i = cycle_parcel[i]
        if np.isnan(load_i):
            y_i = 0
        else:
            y_i = strain_static(load_i, pu_N[i], yc)
        y_N[i] = y_i
        lambda_i = degra_coeff(cycle_i, y_i, b)
        lambda_N[i] = lambda_i
        pu_N[i+1] = pu_N[i] * (1-lambda_i)
    
    return y_N, lambda_N, pu_N



'''
Created on 10/01/2015

@author: mmpe
'''
import numpy as np
from numba.core.decorators import njit


def low_pass(input, delta_t, tau, method=1):

    if isinstance(tau, (int, float)):
        return low_pass_filter(input.astype(np.float64), delta_t, tau)
    else:
        if len(input.shape) == 2:
            return dynamic_low_pass_filter_2d(input.astype(np.float64), delta_t, tau, method)

        else:
            return dynamic_low_pass_filter(input.astype(np.float64), delta_t, tau, method)


def high_pass(input, delta_t, tau):
    return high_pass_filter(input.astype(np.float64), delta_t, tau)


@njit(cache=True)
def low_pass_filter(inp, delta_t, tau):

    output = np.empty_like(inp, dtype=np.float32)
    output[0] = inp[0]

    alpha = delta_t / (tau + delta_t)
    for i in range(1, inp.shape[0]):
        # Same as output[i] = alpha*inp[i]+(1-alpha)*output[i-1]
        output[i] = output[i - 1] + alpha * (inp[i] - output[i - 1])

    return output


@njit(cache=True)
def dynamic_low_pass_filter(inp, delta_t, tau, method=1):

    output = np.empty_like(inp, dtype=np.float32)
    output[0] = inp[0]

    if method == 1:
        alpha = delta_t / (tau + delta_t)
        for i in range(1, inp.shape[0]):
            # Same as output[i] = alpha*inp[i]+(1-alpha)*output[i-1]
            output[i] = output[i - 1] + alpha[i] * (inp[i] - output[i - 1])
    elif method == 2:
        for i in range(1, inp.shape[0]):
            output[i] = (delta_t * (inp[i] + inp[i - 1] - output[i - 1]) +
                         2 * tau[i] * output[i - 1]) / (delta_t + 2 * tau[i])
    elif method == 3:
        for i in range(1, inp.shape[0]):
            output[i] = output[i - 1] * np.exp(-delta_t / tau[i]) + inp[i] * (1 - np.exp(-delta_t / tau[i]))
    return output


@njit(cache=True)
def dynamic_low_pass_filter_2d(inp, delta_t, tau, method=1):

    output = np.empty_like(inp, dtype=np.float32)
    output[0] = inp[0]

    if method == 1:
        alpha = delta_t / (tau + delta_t)
        for i in range(1, inp.shape[0]):
            # Same as output[i] = alpha*inp[i]+(1-alpha)*output[i-1]
            output[i] = output[i - 1] + alpha[i] * (inp[i] - output[i - 1])
    elif method == 2:
        for i in range(1, inp.shape[0]):
            output[i] = (delta_t * (inp[i] + inp[i - 1] - output[i - 1]) +
                         2 * tau[i] * output[i - 1]) / (delta_t + 2 * tau[i])
    elif method == 3:
        for i in range(1, inp.shape[0]):
            output[i] = output[i - 1] * np.exp(-delta_t / tau[i]) + inp[i] * (1 - np.exp(-delta_t / tau[i]))
    return output


@njit(cache=True)
def high_pass_filter(inp, delta_t, tau):
    output = np.empty_like(inp, dtype=np.float32)
    output[0] = inp[0]
    alpha = tau / (tau + delta_t)
    for i in range(1, inp.shape[0]):
        output[i] = alpha * (output[i - 1] + inp[i] - inp[i - 1])

    return output



'''
Created on 17/07/2014

@author: MMPE
'''
from wetb.hawc2.turbulence_file import TurbulenceFile
import os
import unittest
from wetb.hawc2.pc_file import PCFile


import numpy as np

tfp = os.path.join(os.path.dirname(__file__), 'test_files/')  # test file path


class TestPCFile(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)

    # def test_TurbulenceFile(self):
    #    TurbulenceFile(tfp + "turb/mann_l29.4_ae1.00_g3.9_h1_64x8x8_1.000x2.00x2.00_s1001u.turb")

    def test_load_from_htc(self):
        u, v, w = TurbulenceFile.load_from_htc(tfp + "htcfiles/test_turb.htc")
        self.assertEqual(u.data.shape, (512, 64))


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import os

import pandas as pd
from pathlib import Path
import pytest

from wetb.hawc2.tests import test_files
from wetb.hawc2.hawc2_input_writer import HAWC2InputWriter
from wetb.hawc2.htc_file import HTCFile
from wetb.hawc2 import hawc2_input_writer
from tests.run_main import run_module_main


htc_base_dir = os.path.dirname(test_files.__file__) + '/simulation_setup/DTU10MWRef6.0/htc/'


@pytest.fixture
def h2writer():
    """HAWC2InputWriter with DTU 10 MW as base file, used in several tests"""
    htc_base_file = htc_base_dir + 'DTU_10MW_RWT.htc'
    return HAWC2InputWriter(htc_base_file)


def test_hawc2_writer_frompath(tmp_path):
    """Load from a path base, and write to path and string"""
    path = Path(htc_base_dir + 'DTU_10MW_RWT.htc')
    h2writer = HAWC2InputWriter(path)
    h2writer.write(tmp_path / 'test.htc')  # path
    h2writer.write((tmp_path / 'test.htc').as_posix())  # string


def test_hawc2_writer_noname_error(h2writer, tmp_path):
    """Should throw error if set_name is called without name"""
    wsp_lst = [4, 6, 8]
    df = pd.DataFrame({'wind.wsp': wsp_lst, 'Folder': ['tmp'] * 3})
    h2writer.from_pandas(df)
    with pytest.raises(KeyError):  # keyerror in write_all
        h2writer.write_all(tmp_path)


def test_pandas2htc_dottag(h2writer, tmp_path):
    """Load dlc from dict + pandas dataframe (wind speed tag with dot, name), write htc"""
    wsp_lst = [4, 6, 8]
    df = pd.DataFrame({'wind.wsp': wsp_lst, 'Name': ['c1', 'c2', 'c3'], 'Folder': ['tmp'] * 3})
    h2writer.from_pandas(df)
    h2writer.write_all(tmp_path)
    for i, wsp in enumerate(wsp_lst, 1):
        htc_path = tmp_path / ('tmp/c%d.htc' % i)
        htc = HTCFile(htc_path, modelpath=tmp_path.as_posix())
        assert htc.wind.wsp[0] == wsp
        assert htc.simulation.logfile[0] == ('./log/tmp/c%d.log' % i)
        assert htc.output.filename[0] == ('./res/tmp/c%d' % i)


def test_excel2htc(h2writer, tmp_path):
    """Load dlc from Excel (wind speed tag with dot, name), write to htc.
    Both string and pathlib.Path"""
    excel_path = os.path.dirname(test_files.__file__) + '/htc_input_table.xlsx'
    for i in range(2):
        h2writer.from_excel(excel_path if i else Path(excel_path))
        # print(h2writer.contents)
        h2writer.write_all(tmp_path)
        for i, wsp in enumerate([4, 6], 1):
            htc = HTCFile(tmp_path / ('tmp/a%d.htc' % i), modelpath=tmp_path.as_posix())
            assert htc.wind.wsp[0] == wsp
            assert htc.simulation.logfile[0] == ('./log/tmp/a%d.log' % i)
            assert htc.output.filename[0] == ('./res/tmp/a%d' % i)


def test_hawc2_writer_main():
    """Run the example in the hawc2_input_writer module"""
    run_module_main(hawc2_input_writer)


def test_hawc2_writer_custom(tmp_path):
    """Input writer with a custom 'set_'-style attribute"""
    htc_base_file = htc_base_dir + 'DTU_10MW_RWT.htc'

    class MyWriter(HAWC2InputWriter):
        def set_time(self, htc, **kwargs):
            htc.set_time(self.time_start, self.time_start + kwargs['time'])

    myWriter = MyWriter(htc_base_file, time_start=100)
    myWriter.write(tmp_path / 'w1.htc', **{'Name': 'w1', 'wind.wsp': 10, 'time': 600})
    htc = HTCFile(tmp_path / 'w1.htc', modelpath=tmp_path.as_posix())
    assert htc.simulation.time_stop[0] == 700
    assert htc.wind.wsp[0] == 10
    assert htc.output.filename[0] == './res/w1'


def test_CVF2pandas(h2writer):
    """Load dlc with constants, variables, functions format"""
    constants = {'simulation.time_stop': 100}
    variables = {'wind.wsp': [4, 6, 8],
                 'wind.tint': [0.1, 0.15, 0.2]}
    functions = {'Name': lambda x: 'sim_wsp' + str(x['wind.wsp']) + '_ti' + str(x['wind.tint'])}

    h2writer.from_CVF(constants, variables, functions)
    assert len(h2writer.contents) == 9
    assert set(list(h2writer.contents)) == set(['simulation.time_stop', 'wind.wsp', 'wind.tint', 'Name'])



'''
Created on 21/07/2014

@author: mmpe
'''
import numpy as np


def s2matrix(s):
    """Creates a transformation matrix from string

    Parameters
    ----------
    s : string
        x is 
    Returns
    -------
    matrix : array_like
        3x3 transformation matrix             

    Examples
    --------
    >> s2matrix('x,y,z') # identity matrix
    >> s2matrix('x,-z,y) # 90 deg rotation around x

    """
    d = {xyz: v for xyz, v in zip('xyz', np.eye(3))}
    return np.array([eval(xyz, d) for xyz in s.split(",")]).T


def transformation_matrix(angles, xyz):
    """Create Transformation matrix(es)
    !!!Note that the columns of the returned matrix(es) is original(unrotate) xyz-axes in rotated coordinates\n
    !!!Multiplying a this matrix by a vector rotates the vector -angle radians (in right handed terminology).

    Parameters
    ----------
    angles : int, float or array_like
        Angle(s) for rotation matrix(es). \n
        One rotation matrix will be returned for each angle
    xyz : {0,1,2}
        - 0: rotation around x(first) axis\n
        - 1: rotation around y(second) axis\n
        - 2: rotation around z(third) axis\n

    Returns
    -------
    transformation_matrix : array_like, shape = (no_angles, 3,3)
        Rotation matrix(es)
    """
    indexes = [(0, 4, 8, 5, 7), (4, 0, 8, 6, 2), (8, 0, 4, 1, 3)]  # 1, cos,cos,sin,-sin for x,y and z rotation matrix
    if isinstance(angles, (int, float)):
        n = 1
    else:
        n = len(angles)
    m = np.zeros(n * 9, float)
    cosx = np.cos(angles)
    sinx = np.sin(angles)
    m[indexes[xyz][0]::9] = 1
    m[indexes[xyz][1]::9] = cosx
    m[indexes[xyz][2]::9] = cosx
    m[indexes[xyz][3]::9] = sinx
    m[indexes[xyz][4]::9] = -sinx
    return m.reshape(n, 3, 3)


def rotmat(angles, xyz):
    """Create rotation matrix(es)
    !!!Note that the columns of the returned matrix(es) is rotated xyz-axes in original(unrotated) coordinates\n
    Multiplying a this matrix by a vector rotates the vector +angle radians (in right handed terminology).
    Parameters
    ----------
    angles : int, float or array_like
        Angle(s) for rotation matrix(es). \n
        Note that
        One rotation matrix will be returned for each angle
    xyz : {0,1,2}
        - 0: rotation around x(first) axis\n
        - 1: rotation around y(second) axis\n
        - 2: rotation around z(third) axis\n

    Returns
    -------
    transformation_matrix : array_like, shape = (no_angles, 3,3)
        Rotation matrix(es)
    """
    indexes = [(0, 4, 8, 5, 7), (4, 0, 8, 6, 2), (8, 0, 4, 1, 3)]  # 1, cos,cos,sin,-sin for x,y and z rotation matrix
    if isinstance(angles, (int, float)):
        n = 1
    else:
        n = len(angles)
    m = np.zeros(n * 9, float)
    cosx = np.cos(angles)
    sinx = np.sin(angles)
    m[indexes[xyz][0]::9] = 1
    m[indexes[xyz][1]::9] = cosx
    m[indexes[xyz][2]::9] = cosx
    m[indexes[xyz][3]::9] = -sinx
    m[indexes[xyz][4]::9] = sinx
    return m.reshape(n, 3, 3)


def mdot(m1, m2):
    """Multiplication of matrix pairs

    Parameters
    ----------
    m1 : array_like shape = (i, n, m)
        First matrix set, i.e. i matrixes, size n x m\n
        i must be equal to j or 1
    m2 : array_like, shape = (j,m,p)
        Second matrix set, i.e. j matrixes, size m x p\n
        j must be equal to i or 1

    Returns
    -------
    mdot : array_like, shape(max(i,j),n,p)
        Matrix products
    """
    if m1.shape[0] == 1 and m2.shape[0] == 1:
        return np.array([np.dot(m1[0], m2[0])])
    elif m1.shape[0] > 1 and m2.shape[0] == 1:
        mprod = np.empty_like(m1)
        for i in range(m1.shape[0]):
            mprod[i, :] = np.dot(m1[i, :], m2[0])
    elif m1.shape[0] == 1 and m2.shape[0] > 1:
        mprod = np.empty_like(m2)
        for i in range(m2.shape[0]):
            mprod[i, :] = np.dot(m1[0], m2[i, :])
    elif m1.shape[0] > 1 and m2.shape[0] > 1 and m1.shape[0] == m2.shape[0]:
        mprod = np.empty_like(m1)
        for i in range(m1.shape[0]):
            mprod[i, :] = np.dot(m1[i, :], m2[i, :])
    else:
        raise Exception("m1 and m2 must have same first dimension or one of the matrices must have first dimension 1")
    return mprod


def dots(rotmats, v):
    """Rotate vector(s) v by rotation matrix(es)

    Parameters
    ----------
    rotmats : array_like, shape=(i,3,3) or list of array_like, shape=(i,3,3)
        if list of array_like, rotation matrixes are recursively reduced by multiplication\n
        i must be 1 or equal to j
    vector : array_like, shape=(j,3)
        vectors to rotate

    Returns
    -------
    dots : array_like, shape=(j,3)
    """
    if isinstance(rotmats, list):
        if len(rotmats) > 1:
            rotmats_red = [mdot(rotmats[0], rotmats[1])]
            rotmats_red.extend(rotmats[2:])
            return dots(rotmats_red, v)
        else:
            m = rotmats[0]
    else:
        m = rotmats
    if len(v.shape) == 1:
        v = np.array([v]).T

    if m.shape[0] == 1:
        return np.dot(m[0], v)
    else:
        if m.shape[0] != v.shape[1]:
            raise Exception("m must have same dimension as v has number of columns")
        res = np.zeros_like(v, dtype=float)
        for i in range(v.shape[1]):
            res[:, i] = np.dot(m[i], v[:, i])
        return res


def rotate(rotmats, v):
    global x, y, z
    v = np.array(v)
    if isinstance(rotmats, (list, tuple)):
        if len(rotmats) > 1:
            rotmats_red = [mdot(rotmats[0], rotmats[1])]
            rotmats_red.extend(rotmats[2:])
            return rotate(rotmats_red, v)
        else:
            rotmats = rotmats[0]
    assert rotmats.shape[-1] == rotmats.shape[-2] == v.shape[-1] == 3, "rotmats is %d x %x and v is i x %d, but must be 3" % (
        rotmats.shape[-1], rotmats.shape[-2], v.shape[-1])

#    if isinstance(v, tuple):
#        v = np.array([v])
    n = (1, v.shape[0])[len(v.shape) == 2]

    if rotmats.shape[0] == 1:
        return np.dot(rotmats[0], v.T).T
    else:
        if v.shape[0] != rotmats.shape[0]:
            raise Exception("V and rotmats must have same first dimension")
        v = v.T
        v_rot = np.zeros_like(v, dtype=float)
        for i in range(n):
            v_rot[:, i] = np.dot(rotmats[i], v[:, i])
        return v_rot.T


def rotate_x(v, angle):
    """Rotate vector(s) around x axis

    Parameters
    ---------
    v : array_like, shape=(3,) or shape=(1-N,3)
        Vector(s) to rotate
    angle : int, float or array_like shape=(1,) or shape=(N,)
        Angle(s) [rad] to rotate

    Returns
    -------
    y : array_like
        Rotated vector(s)
    """
    return _rotate(v, angle, lambda x, y, z, cos, sin: [x, cos * y - sin * z, sin * y + cos * z])


def rotate_y(v, angle):
    """Rotate vector(s) around y axis

    Parameters
    ---------
    v : array_like, shape=(3,) or shape=(1-N,3)
        Vector(s) to rotate
    angle : int, float or array_like shape=(1,) or shape=(N,)
        Angle(s) [rad] to rotate

    Returns
    -------
    y : array_like
        Rotated vector(s)
    """
    return _rotate(v, angle, lambda x, y, z, cos, sin: [cos * x + sin * z, y, -sin * x + cos * z])


def rotate_z(v, angle):
    """Rotate vector(s) around z axis

    Parameters
    ---------
    v : array_like, shape=(3,) or shape=(1-N,3)
        Vector(s) to rotate
    angle : int, float or array_like shape=(1,) or shape=(N,)
        Angle(s) [rad] to rotate

    Returns
    -------
    y : array_like
        Rotated vector(s)
    """
    return _rotate(v, angle, lambda x, y, z, cos, sin: [cos * x - sin * y, sin * x + cos * y, z])


def _rotate(v, angle, rotfunc):
    angle = np.atleast_1d(angle)
    cos, sin = np.cos(angle), np.sin(angle)
    v = np.array(v)
    if len(v.shape) == 1:
        assert angle.shape[0] == 1
        assert v.shape[0] == 3
        return np.array(rotfunc(v[0], v[1], v[2], cos[0], sin[0]), dtype=float).T
    else:
        assert angle.shape[0] == 1 or angle.shape[0] == v.shape[0]
        assert v.shape[1] == 3
        return np.array(rotfunc(v[:, 0], v[:, 1], v[:, 2], cos, sin), dtype=float).T


#=======================================================================================================================
# Conversions
#=======================================================================================================================
# http://www.euclideanspace.com/maths/geometry/rotations/conversions/
# - axis: [x,y,z]*angle_deg
# - axis_angle: [x,y,z,angle_rad]
# - Quaternion: [qw,qx,qy,qz]
# - Matrix 3x3 transformation matrix

def norm(vector):
    return np.sqrt(np.sum(np.asarray(vector)**2))


#=======================================================================================================================
# axis to ...
#=======================================================================================================================
def axis2axis_angle(axis):
    """
    deg : boolean
        if True, axis length is assumed to be angle in deg 
    """
    axis = np.asarray(axis)
    angle = np.sqrt(((axis**2).sum()))
    return np.r_[axis / np.sqrt((axis**2).sum()), angle]


def axis2matrix(axis, deg=False):
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/angleToMatrix/index.htm
    axis = np.asarray(axis)
    angle = np.sqrt(((axis**2).sum()))
    if deg:
        angle = np.deg2rad(angle)
    if angle == 0:
        return np.eye(3)
    x, y, z = xyz = axis / np.sqrt((axis**2).sum())

    c, s = np.cos(angle), np.sin(angle)
    t = 1 - c

    return np.array([[t * x * x + c, t * x * y - z * s, t * x * z + y * s],
                     [t * x * y + z * s, t * y * y + c, t * y * z - x * s],
                     [t * x * z - y * s, t * y * z + x * s, t * z * z + c]])
    # alternative implementation
    #asym = np.array([[0, -z, y], [z, 0, -x], [-y, x, 0]])
    # return c * np.eye(3, 3) + s * asym + (1 - c) * xyz[np.newaxis] * xyz[:, np.newaxis]

def axis2quaternion(axis):

    axis = np.asarray(axis)
    q = np.zeros([axis.shape[0], 4])

    if len(axis.shape) > 1:
        yaw = axis[:,2]
        pitch = axis[:,1]
        roll = axis[:,0]
    else:
        yaw = axis[2]
        pitch = axis[1]
        roll = axis[0]

    cy = np.cos(yaw * 0.5)
    sy = np.sin(yaw * 0.5)
    cp = np.cos(pitch * 0.5)
    sp = np.sin(pitch * 0.5)
    cr = np.cos(roll * 0.5)
    sr = np.sin(roll * 0.5)

    qw = cr * cp * cy + sr * sp * sy
    qx = sr * cp * cy - cr * sp * sy
    qy = cr * sp * cy + sr * cp * sy
    qz = cr * cp * sy - sr * sp * cy

    q[:, 0] = qw
    q[:, 1] = qx
    q[:, 2] = qy
    q[:, 3] = qz

    return q
#=======================================================================================================================
# # axis_angle to ...
#=======================================================================================================================


def axis_angle2axis(axis_angle):
    x, y, z, angle = axis_angle
    return np.array([x, y, z]) * angle


def axis_angle2quaternion(axis_angle, deg=False):
    x, y, z, angle = axis_angle
    if deg:
        angle = np.deg2rad(angle)
    s = np.sin(angle / 2)
    x = x * s
    y = y * s
    z = z * s
    w = np.cos(angle / 2)
    return w, x, y, z


#=======================================================================================================================
# # quaternion to ...
#=======================================================================================================================

def quaternion2axis_angle(quaternion, deg=False):
    qw, qx, qy, qz = quaternion / norm(quaternion)
    angle = 2 * np.arccos(qw)
    if deg:
        angle = np.rad2deg(angle)
    t = np.sqrt(1 - qw**2)
    x = qx / t
    y = qy / t
    z = qz / t
    return x, y, z, angle


def quaternion2matrix(quaternion):
    q = quaternion / np.linalg.norm(quaternion, axis=0)
    qw, qx, qy, qz = q
    sqw, sqx, sqy, sqz = q**2

    qxw = qx * qw
    qxy = qx * qy
    qxz = qx * qz
    qyw = qy * qw
    qyz = qy * qz
    qzw = qz * qw

    return np.array([[sqx - sqy - sqz + sqw, 2.0 * (qxy - qzw), 2.0 * (qxz + qyw)],
                     [2.0 * (qxy + qzw), -sqx + sqy - sqz + sqw, 2.0 * (qyz - qxw)],
                     [2.0 * (qxz - qyw), 2.0 * (qyz + qxw), -sqx - sqy + sqz + sqw]])

#=======================================================================================================================
# Matrix to ...
#=======================================================================================================================


def matrix2quaternion(matrix):
    # method from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
    sqrt = np.sqrt
    (m00, m01, m02), (m10, m11, m12), (m20, m21, m22) = matrix
    tr = m00 + m11 + m22

    if (tr > 0):
        S = sqrt(tr + 1.0) * 2  # // S=4*qw
        qw = 0.25 * S
        qx = (m21 - m12) / S
        qy = (m02 - m20) / S
        qz = (m10 - m01) / S
    elif ((m00 > m11) and (m00 > m22)):
        S = sqrt(1.0 + m00 - m11 - m22) * 2  # // S=4*qx
        qw = (m21 - m12) / S
        qx = 0.25 * S
        qy = (m01 + m10) / S
        qz = (m02 + m20) / S
    elif (m11 > m22):
        S = sqrt(1.0 + m11 - m00 - m22) * 2  # // S=4*qy
        qw = (m02 - m20) / S
        qx = (m01 + m10) / S
        qy = 0.25 * S
        qz = (m12 + m21) / S
    else:
        S = sqrt(1.0 + m22 - m00 - m11) * 2  # // S=4*qz
        qw = (m10 - m01) / S
        qx = (m02 + m20) / S
        qy = (m12 + m21) / S
        qz = 0.25 * S
    e = np.array([qw, qx, qy, qz])
    return e / np.sqrt(np.sum(e**2))


def matrix2axis_angle(matrix, deg=False):
    return quaternion2axis_angle(matrix2quaternion(matrix), deg)


def matrix2axis(matrix, deg=False):
    return axis_angle2axis(matrix2axis_angle(matrix, deg))



import itertools
import os
from pathlib import Path
import warnings

from tqdm import tqdm
import jinja2
import pandas as pd
from pandas.core.base import PandasObject
import numpy as np

from wetb.hawc2.htc_file import HTCFile


class HAWC2InputWriter(object):
    """Load a dlc and write the corresponding HAWC2 input files.

    Parameters
    -----------
    base_htc_file : str or Pathlib.path
        Path to base htc file from which the other htc files are created.

    Notes
    ------
    You can load the dlc keywords and parameters using the `from_` methods presented
    below. Any methods of the form `set_<tag>`(htc, **kwargs), where `<tag>` is one
    of one of the keywords in the dlc, will be called when writing htc files.

    You can write the htc files using the `write` or `write_all` methods presented below.
    The `Name` tag must be defined when writing htc files.

    Necessary/protected tags
    -------------------------
    Name : str
        The name of the htc file (without extension) for each simulation. This will
        be used in a `set_Name` call to the htc file, which will overwrite the
        `simulation.logfile` and `output.filename` lines.
    """

    def __init__(self, base_htc_file, **kwargs):
        self.base_htc_file = base_htc_file
        self.contents = None
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __call__(self, path, **kwargs):
        """Write single htc file if writer is called directly"""
        return self.write(path, **kwargs)

    def from_pandas(self, dataFrame):
        """Load dlc contents from a PandasObject.

        Parameters
        ----------
        dataFrame : Pandas.DataFrame or PandasObject
            Dataframe of contents.

        Returns
        -------
        self : HAWC2InputWriter
            Instance of self to enable chaining.
        """
        if not isinstance(dataFrame, PandasObject) and hasattr(dataFrame, 'to_pandas'):
            dataFrame = dataFrame.to_pandas()
        self.contents = dataFrame
        return self

    def from_excel(self, excel_file):
        """Load dlc contents from an excel file.

        Parameters
        ----------
        excel_file : str or Pathlib.Path
            Path to excel file.

        Returns
        -------
        self : HAWC2InputWriter
            Instance of self to enable chaining.
        """

        self.contents = pd.read_excel(excel_file, keep_default_na=False)
        if 'Unnamed: 0' in self.contents:
            self.contents = self.contents.drop(columns=['Unnamed: 0'])
        return self

    def from_CVF(self, constants, variables, functionals):
        """Load dlc contents from dictionaries of constants, variables and functionals.

        Parameters
        ----------
        constants : dict
            (content, value) pairs for which the value remains constant for all
            generated htc files. E.g.: {'time_stop': 700}.
        variables : dict of lists
            (content, list) pairs for which all combinations of the listed values are
            generated in the htc files. E.g., {'wsp': [4, 6, 8]}.
        functionals : dict of functions
            (content, function) pairs for which the content is dependent on the
            constants and variables contents. E.g., {'Name': lambda x: 'wsp' + str(x['wsp'])}.

        Returns
        -------
        self : HAWC2InputWriter
            Instance of self to enable chaining.
        """
        attributes = list(constants.keys()) + list(variables.keys()) + list(functionals.keys())
        contents = []

        for var in itertools.product(*list(variables.values())):
            this_dict = dict(constants)
            var_dict = dict(zip(variables.keys(), var))
            this_dict.update(var_dict)

            for key, func in functionals.items():
                this_dict[key] = func(this_dict)

            contents.append(list(this_dict.values()))

        self.contents = pd.DataFrame(contents, columns=attributes)

        return self

    def set_Name(self, htc, **kwargs):
        """Update the filename in simulation and output blocks of an htc file.

        Notes
        -----
        This function is called during the `write` and `write_all` methods, thereby
        overwriting the logfile and output filename. If you would like to specify a
        subfolder within the `htc` and `res` directories where your simulations should
        be placed, pass in the 'Folder' keyword into `kwargs`.

        Parameters
        ----------
        htc : wetb.hawc2.htc_file.HTCFile
            The htc file object to be modified.
        name : str
            The filename to be used in the logfile and simulation file (without
            extension).
        **kwargs
            Keyword arguments. Must include keyword 'Name'. May optionally include
            'Folder' to define a subfolder in your `htc` and `res` directories where
            files will be stored.
        """
        htc.set_name(kwargs['Name'], subfolder=kwargs.get('Folder', ''))

    def write(self, path, **kwargs):
        """Write a single htc file.

        Notes
        -----
        This function works both with the tagless and jinja file-writing systems. Any
        methods of the form `set_<tag>` are called during file writing.

        Parameters
        ----------
        path : str or pathlib.Path
            The path to save the htc file to.
        **kwargs : pd.DataFrame, pd.Series or dict
            Keyword arguments. The tags to update/replace in the file.
        """
        htc = HTCFile(self.base_htc_file, jinja_tags=kwargs)
        for k, v in kwargs.items():
            k = k.replace('/', '.')
            if '.' in k:  # directly replace tags like "wind.wsp"
                command = k.split(".")[-1]
                subsections = k.split(".")[:-1]
                section = htc
                for subsection in subsections:
                    section = section.add_section(subsection)
                if command in section:
                    section[command].values = np.atleast_1d(v).tolist()
                else:
                    section[command] = v
            else:  # otherwise, use the "set_" attribute
                if hasattr(self, 'set_%s' % k):
                    getattr(self, 'set_%s' % k)(htc, **kwargs)
        htc.save(path)

    def write_all(self, out_dir):
        """Write all htc files for the dlc.

        Notes
        -----
        This function works both with the tagless and jinja file-writing systems. Any
        methods of the form `set_<tag>` are called during each htc write. DLC contents
        must contain a 'Name' column.

        Parameters
        ----------
        out_dir : str or pathlib.Path
            The directory in which the htc files should be written.
        """
        try:
            self.contents['Name']
        except KeyError:
            raise KeyError('"Name" not given in dlc contents! Cannot write files.')

        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)

        N = len(self.contents)

        print(f'Generating {N} htc files in directory: {out_dir}')

        with tqdm(self.contents.iterrows(), total=N) as bar:
            for _, row in bar:
                if 'Folder' in row and row.Folder:
                    path = out_dir / row.Folder
                else:
                    path = out_dir
                self.write(path / (row.Name + '.htc'), **row.to_dict())


class JinjaWriter(HAWC2InputWriter):
    def __init__(self, base_htc_file, **kwargs):
        """DEPRECATED!!!
        Subclass of the HAWC2InputWriter object. Generates htc files using contents compatible with jinja2.
        """
        warnings.warn('The JinjaWriter is deprecated! Please switch to the HAWC2InputWriter',
                      DeprecationWarning)
        HAWC2InputWriter.__init__(self, base_htc_file, **kwargs)

    def write(self, out_fn, **kwargs):
        params = pd.Series(kwargs)
        with open(self.base_htc_file) as f:
            template = jinja2.Template(f.read())

            out = template.render(params)

            with open(out_fn, 'w') as f:
                f.write(out)


def main():
    if __name__ == '__main__':
        from wetb.hawc2.tests import test_files

        # example of subclassing with custom set_<tag> method
        path = os.path.dirname(test_files.__file__) + '/simulation_setup/DTU10MWRef6.0/'
        base_htc_file = path + 'htc/DTU_10MW_RWT.htc'

        class MyWriter(HAWC2InputWriter):
            def set_time(self, htc, **kwargs):
                htc.set_time(self.time_start, self.time_start + kwargs['time'])

        myWriter = MyWriter(base_htc_file, time_start=100)
        myWriter('tmp/t1.htc', Name='t1', **{'time': 600, 'wind.wsp': 10})

        # example with constants, variables, and functionals
        Constants = {
            'time_stop': 100,
        }

        Variables = {
            'wsp': [4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24],
            'tint': [0.05, 0.10, 0.15],

        }

        Functionals = {
            'Name': lambda x: 'dtu10mw_wsp' + str(x['wsp']) + '_ti' + str(x['tint']),
        }

        htc_template_fn = os.path.dirname(test_files.__file__) + \
            '/simulation_setup/DTU10MWRef6.0/htc/DTU_10MW_RWT.htc.j2'
        writer = HAWC2InputWriter(htc_template_fn)
        writer.from_CVF(Constants, Variables, Functionals)
        print(writer.contents)

        # write all htc files to folder 'tmp'
        writer.write_all('tmp')


main()



# -*- coding: utf-8 -*-
"""
Created on Mon Aug  8 10:22:49 2016

@author: dave
"""
# standard python library
import os
from os.path import join as pjoin
import zipfile
import copy
import tarfile
import glob
import shutil
import tempfile

import numpy as np
import pandas as pd
#from tqdm import tqdm

from wetb.prepost.Simulations import Cases
from wetb.prepost import misc


def create_chunks_htc_pbs(cases, sort_by_values=['[Windspeed]'], ppn=20, i0=0,
                          nr_procs_series=9, queue='workq', compress=False,
                          walltime='24:00:00', chunks_dir='zip-chunks-jess',
                          wine_arch='win32', wine_prefix='.wine32',
                          pyenv_cmd='source /home/python/miniconda3/bin/activate',
                          pyenv='py36-wetb', prelude='', ppn_pbs=20):
    """Group a large number of simulations htc and pbs launch scripts into
    different zip files so we can run them with find+xargs on various nodes.
    """

    def chunker(seq, size):
        # for DataFrames you can also use groupby, as taken from:
        # http://stackoverflow.com/a/25703030/3156685
        # for k,g in df.groupby(np.arange(len(df))//10)
        # but this approach is faster, see also:
        # http://stackoverflow.com/a/25701576/3156685
        # http://stackoverflow.com/a/434328/3156685
        return (seq[pos:pos + size] for pos in range(0, len(seq), size))

    def make_zip_chunks(df, ii, sim_id, run_dir, model_zip):
        """Create zip cunks and also create an index
        """

        # create a new zip file, give index of the first element. THis is
        # quasi random due to the sorting we applied earlier
#        ii = df.index[0]
        rpl = (sim_id, ii)
        fname = os.path.join(run_dir, chunks_dir, '%s_chnk_%05i' % rpl)
        zf = zipfile.ZipFile(fname+'.zip', 'w', compression=zipfile.ZIP_STORED)

        # start with appending the base model zip file
        fname_model = os.path.join(run_dir, model_zip)
        with zipfile.ZipFile(fname_model, 'r') as zf_model:
            for n in zf_model.namelist():
                # FIXME: might be duplicates when creating dirs from dirtags
                zf.writestr(n, zf_model.open(n).read())

        # create all necessary directories in the zip file
        dirtags = ['[htc_dir]', '[res_dir]','[log_dir]','[animation_dir]',
                   '[pbs_in_dir]', '[eigenfreq_dir]','[turb_dir]','[micro_dir]',
                   '[meander_dir]','[hydro_dir]', '[mooring_dir]',
                   '[pbs_in_dir]', '[pbs_out_dir]']
        dirnames = []
        for tag in dirtags:
            for dirname in set(df[tag].unique().tolist()):
                if not dirname or dirname.lower() not in ['false', 'none', 0]:
                    dirnames.append(dirname)

        namelist = set(zf.namelist())
        for dirname in set(dirnames):
            if dirname != 0 and pjoin(dirname, '') not in namelist:
                # FIXME: might have duplicates from the base model zip file
                zf.write('.', arcname=os.path.join(dirname, '.'))

        # and the post-processing data
        # FIXME: do not use hard coded paths!
        zf.write('.', arcname='prepost-data/')

        # HTC files
        df_src = df['[run_dir]'] + df['[htc_dir]'] + df['[case_id]']
        df_dst = df['[htc_dir]'] + df['[case_id]']
        # create an index so given the htc file, we can find the chunk nr
        df_index = pd.DataFrame(index=df['[case_id]'].copy(),
                                columns=['chnk_nr', 'name'])
        df_index['chnk_nr'] = ii
        df_index['name'] = os.path.join(chunks_dir, '%s_chnk_%05i' % rpl)
        # Since df_src and df_dst are already Series, iterating is fast an it
        # is slower to first convert to a list
        for src, dst_rel in zip(df_src, df_dst):
            zf.write(src+'.htc', arcname=dst_rel+'.htc')

        # PBS files
        df_src = df['[run_dir]'] + df['[pbs_in_dir]'] + df['[case_id]']
        df_dst = df['[pbs_in_dir]'] + df['[case_id]']
        # Since df_src and df_dst are already Series, iterating is fast an it
        # is slower to first convert to a list
        for src, dst_rel in zip(df_src, df_dst):
            zf.write(src+'.p', arcname=dst_rel+'.p')

        # copy and rename input files with given versioned name to the
        # all files that will have to be renamed to their non-changeable
        # default file name.
        # this is a bit more tricky since unique() will not work on list items
        copyto_files_tmp = df['[copyto_files]'].astype(str)
        copyto_files = []
        # cycle through the unique elements
        for k in set(copyto_files_tmp):
            # k is of form: 'some/file.txt**another/file1.txt
            if len(k) < 1:
                continue
            # note that *;* is done in df_dict_check_datatypes()
            copyto_files.extend(k.split('*;*'))
        # we might still have non unique elements
        copyto_files = set(copyto_files)
        for copyto_file, dst_rel in zip(copyto_files, df_dst):
            src = os.path.join(run_dir, copyto_file)
            # make dir if it does not exist
            namelist = set(zf.namelist())
            # write an empty directory if applicable, make sure ends with /
            copyto_file_folder = pjoin(os.path.dirname(copyto_file), '')
            if len(copyto_file_folder) > 0:
                if copyto_file_folder not in namelist:
                    zf.write('.', arcname=copyto_file_folder)
            # if we have a wildcard, copy all files accordingly
            for fname in glob.glob(copyto_file, recursive=True):
                print(fname)
                if copyto_file not in namelist:
                    zf.write(fname, arcname=fname)
        zf.close()

        return fname, df_index

    pbs_tmplate = "\n"
    pbs_tmplate += "### Standard Output\n"
    pbs_tmplate += "#PBS -N [job_name]\n"
    pbs_tmplate += "#PBS -o [std_out]\n"
    pbs_tmplate += "### Standard Error\n"
    pbs_tmplate += "#PBS -e [std_err]\n"
    pbs_tmplate += "#PBS -W umask=[umask]\n"
    pbs_tmplate += "### Maximum wallclock time format HOURS:MINUTES:SECONDS\n"
    pbs_tmplate += "#PBS -l walltime=[walltime]\n"
    pbs_tmplate += "#PBS -l nodes=[nodes]:ppn=[ppn_pbs]\n"
    pbs_tmplate += "### Queue name\n"
    pbs_tmplate += "#PBS -q [queue]\n"
    pbs_tmplate += "\n"

    # FIXME: this causes troubles on CI runner for the tests (line endings?)
#     pbs_tmplate = """
# ### Standard Output
# #PBS -N [job_name]
# #PBS -o [std_out]
# ### Standard Error
# #PBS -e [std_err]
# #PBS -W umask=[umask]
# ### Maximum wallclock time format HOURS:MINUTES:SECONDS
# #PBS -l walltime=[walltime]
# #PBS -l nodes=[nodes]:ppn=[ppn]
# ### Queue name
# #PBS -q [queue]

# """

    def make_pbs_chunks(df, ii, sim_id, run_dir, model_zip, compress=False,
                        wine_arch='win32', wine_prefix='.wine32'):
        """Create a PBS that:
            * copies all required files (zip chunk) to scratch disk
            * copies all required turbulence files to scratch disk
            * runs everything with find+xargs
            * copies back what's need to mimer
            """
#        ii = df.index[0]
        cmd_find = '/home/MET/sysalt/bin/find'
        cmd_xargs = '/home/MET/sysalt/bin/xargs'
        jobid = '%s_chnk_%05i' % (sim_id, ii)

        # sanitize wine_prefix
        wine_prefix = misc.sanitize_wine_prefix(wine_prefix)

        wineparam = (wine_arch, wine_prefix)

        pbase = os.path.join('/scratch','$USER', '$PBS_JOBID', '')
        post_dir_base = post_dir.split(sim_id)[1]
        if post_dir_base[0] == os.path.sep:
            post_dir_base = post_dir_base[1:]

        # FIXME: commonpath was only added in Python 3.5, but CI runner is old
        try:
            compath = os.path.commonpath
        except AttributeError:
            compath = os.path.commonprefix

        pbs_in_base = compath(df['[pbs_in_dir]'].unique().tolist())
        pbs_in_base = os.path.join(pbs_in_base, '')
        htc_base = compath(df['[htc_dir]'].unique().tolist())
        htc_base = os.path.join(htc_base, '')
        res_base = compath(df['[res_dir]'].unique().tolist())
        res_base = os.path.join(res_base, '')
        log_base = compath(df['[log_dir]'].unique().tolist())
        log_base = os.path.join(log_base, '')

        # =====================================================================
        # PBS HEADER
        pbs = copy.copy(pbs_tmplate)
        pbs = pbs.replace('[job_name]', jobid)
        pbs = pbs.replace('[std_out]', './pbs_out_chunks/%s.out' % jobid)
        pbs = pbs.replace('[std_err]', './pbs_out_chunks/%s.err' % jobid)
        pbs = pbs.replace('[umask]', '0003')
        pbs = pbs.replace('[walltime]', walltime)
        pbs = pbs.replace('[nodes]', str(1)) # only one node for the time being
        pbs = pbs.replace('[ppn_pbs]', str(ppn_pbs))
        pbs = pbs.replace('[queue]', queue)
        pbs += '\necho "%s"\n' % ('-'*70)

        # run prelude code
        # =====================================================================
        pbs += prelude

        # =====================================================================
        # activate the python environment
        if pyenv is not None:
            pbs += 'echo "activate python environment %s"\n' % pyenv
            rpl = (pyenv_cmd, pyenv)
            pbs += '%s %s\n' % rpl
            # sometimes activating an environment fails due to a FileExistsError
            # is this because it is activated at the same time on another node?
            # check twice if the environment got activated for real,
            # but only do so for /home/python/miniconda
            if pyenv_cmd.find('miniconda') > -1:
                pbs += 'echo "CHECK 2x IF %s IS ACTIVE, IF NOT TRY AGAIN"\n' % pyenv
                pbs += 'CMD=\"from distutils.sysconfig import get_python_lib;'
                pbs += 'print (get_python_lib().find(\'/usr/lib/python\'))"\n'
                pbs += 'ACTIVATED=`python -c "$CMD"`\n'
                pbs += 'if [ $ACTIVATED -eq 0 ]; then %s %s;fi\n' % rpl
                pbs += 'ACTIVATED=`python -c "$CMD"`\n'
                pbs += 'if [ $ACTIVATED -eq 0 ]; then %s %s;fi\n' % rpl

        # =====================================================================
        # create all necessary directories at CPU_NR dirs
        # browse to scratch directory
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % pbase
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n\n'
        pbs += 'echo "create CPU directories on the scratch disk"\n'
        pbs += 'mkdir -p "%s"\n' % os.path.join(pbase, sim_id, '')
        for k in range(ppn):
            pbs += 'mkdir -p "%s"\n' % os.path.join(pbase, '%i' % k, '')

        # =====================================================================
        # get the zip-chunk file from the PBS_O_WORKDIR
        pbs += '\n'
        pbs += 'echo "%s"\n' % ('-'*70)
        pbs += 'cd $PBS_O_WORKDIR\n'
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "get the zip-chunk file from the PBS_O_WORKDIR"\n'
        # copy the relevant zip chunk file to the scratch main directory
        rpl = (os.path.join(chunks_dir, jobid), os.path.join(pbase, ''))
        pbs += 'cp "%s.zip" "%s"\n' % rpl

        # =====================================================================
        # unzip to all cpu dirs
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % pbase
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "unzip chunk, create dirs in cpu and sim_id folders"\n'
        # unzip chunk, this contains all relevant folders already, and also
        # contains files defined in [copyto_files]
        for k in list(range(ppn)) + [sim_id]:
            dst = os.path.join('%s' % k, '.')
            pbs += '/usr/bin/unzip "%s" -d "%s" >> /dev/null\n' % (jobid+'.zip', dst)

        # =====================================================================
        # create all turb_db directories
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % os.path.join(pbase, sim_id, '')
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "create turb_db directories"\n'
        turb_db_tags = ['[turb_db_dir]', '[meander_db_dir]', '[micro_db_dir]']
        turb_db_dirs = []
        for tag in turb_db_tags:
            for dirname in set(df[tag].unique().tolist()):
                dirname_s = str(dirname).replace('.', '').replace('/', '')
                if dirname_s.lower() not in ['false', 'none', '0']:
                    turb_db_dirs.append(dirname)
        turb_db_dirs = set(turb_db_dirs)
        # create all turb dirs
        for dirname in sorted(turb_db_dirs):
            pbs += 'mkdir -p "%s"\n' % os.path.join(dirname, '')

        # =====================================================================
        # copy required turbulence from db_dir to scratch/db_dirs
        # turb_db_dir might not be set, same for turb_base_name, for those
        # cases we do not need to copy anything from the database to the node
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "$PBS_O_WORKDIR"\n'
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        base_name_tags = ['[turb_base_name]', '[meander_base_name]',
                          '[micro_base_name]']
        for db, base_name in zip(turb_db_tags, base_name_tags):
            turb_db_dirs = df[db] + df[base_name]
            # When set to None, the DataFrame will have text as None
            # FIXME: CI runner has and old pandas version (v0.14.1)
            try:
                p1 = turb_db_dirs.str.find('None')==-1
                p2 = turb_db_dirs.str.find('none')==-1
                p3 = turb_db_dirs.str.find('false')==-1
                p4 = turb_db_dirs.str.find('False')==-1
#                p4 = turb_db_dirs.str.find('0')==-1
                turb_db_src = turb_db_dirs[p1 & p2 & p3 & p4]
            except AttributeError:
                # and findall returns list with the search str occuring as
                # many times as found in the str...
                # sel should be True if str does NOT occur in turb_db_dirs
                # meaning if findall returns empty list
                sel = [True]*len(turb_db_dirs)
                for val in ['false', 'none', 'None', 'False']:
                    findall = turb_db_dirs.str.findall(val).tolist()
                    # len==0 if nothing has been found
                    sel_ = [True if len(k)==0 else False for k in findall]
                    # merge with other search results, none of the elements
                    # should occur
                    sel = [True if k and kk else False for k, kk in zip(sel, sel_)]
                turb_db_src = turb_db_dirs[sel]
            pbs += '\n'
            pbs += '# copy to scratch db directory for %s, %s\n' % (db, base_name)
            for k in sorted(turb_db_src.unique()):
                dst = os.path.dirname(os.path.join(pbase, sim_id, k))
                # globbing doesn't work in either single- or double-quotes.
                # However, you can interpolate globbing with double-quoted strings
                # https://unix.stackexchange.com/a/67761/163108
                pbs += 'cp "%s"*.bin "%s"\n' % (k, os.path.join(dst, '.'))

        # =====================================================================
        # to be safe, create all turb dirs in the cpu dirs
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % os.path.join(pbase, '')
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "create turb directories in CPU dirs"\n'
        turb_dir_tags = ['[turb_dir]', '[meander_dir]', '[micro_dir]']
        turb_dirs = []
        for tag in turb_dir_tags:
            for dirname in set(df[tag].unique().tolist()):
                dirname_s = str(dirname).replace('.', '').replace('/', '')
                if dirname_s.lower() not in ['false', 'none', '0']:
                    turb_dirs.append(dirname)
        turb_dirs = sorted(set(turb_dirs))
        for k in list(range(ppn)):
            for dirname in turb_dirs:
                pbs += 'mkdir -p "%s"\n' % os.path.join(str(k), dirname, '')

        # =====================================================================
        # symlink everything from the turb_db_dir to the cpu/turb_dir
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % os.path.join(pbase, sim_id, '')
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "Link all turb files into CPU dirs"\n'
        for db_dir, turb_dir in zip(turb_db_tags, turb_dir_tags):
            # FIXME: this needs to be written nicer. We should be able to
            # select from df non-defined values so we can exclude them
            # now it seems they are either None, False or 0 in either
            # boolean, str or int formats
            nogo = ['false', 'none', '0']
            try:
                symlink_dirs = df[db_dir] + '_*_'
                symlink_dirs = symlink_dirs + df[turb_dir]
            except:
                continue
            for symlink in symlink_dirs.unique().tolist():
                db_dir, turb_dir = symlink.split('_*_')
                db_dir_s = db_dir.replace('.', '').replace('/', '').lower()
                turb_dir_s = turb_dir.replace('.', '').replace('/', '').lower()
                if db_dir_s in nogo or turb_dir_s in nogo:
                    continue
                db_dir_abs = os.path.join(pbase, sim_id, db_dir, '')
                for k in list(range(ppn)):
                    turb_dir_abs = os.path.join(pbase, str(k), turb_dir, '')
                    rpl = (db_dir_abs, turb_dir_abs)
                    pbs += 'find "%s" -iname "*.bin" -exec ln -s {} "%s" \\;\n' % rpl

        # copy all from scratch/turb_db to cpu/turb
        # turb_dir_base = os.path.join(compath(list(turb_dirs)), '')
        # pbs += '\necho "%s"\n' % ('-'*70)
        # pbs += 'cd %s\n' % os.path.join(pbase, sim_id, '')
        # pbs += "echo 'current working directory:'\n"
        # pbs += 'pwd\n'
        # pbs += 'echo "Link all turb files into CPU dirs"\n'
        # for k in range(ppn):
        #     rpl = (os.path.relpath(os.path.join(sim_id, turb_dir_base)), k)
        #     pbs += 'find %s -iname "*.bin" -exec cp {} %s/{} \\;\n' % rpl

        # =====================================================================
        # finally we can run find+xargs!!!
        pbs += '\n'
        pbs += 'echo "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % pbase
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "START RUNNING JOBS IN find+xargs MODE"\n'
        if wine_arch!=None or wine_prefix!=None:
            pbs += 'WINEARCH="%s" WINEPREFIX="%s" winefix\n' % wineparam
        pbs += '# run all the PBS *.p files in find+xargs mode\n'
        pbs += 'echo "following cases will be run from following path:"\n'
        pbs += 'echo "%s"\n' % (os.path.join(sim_id, pbs_in_base))
        pbs += 'export LAUNCH_PBS_MODE=false\n'
        rpl = (cmd_find, os.path.join(sim_id, pbs_in_base))
        pbs += "%s '%s' -type f -name '*.p' | sort -z\n" % rpl
        pbs += '\n'
        pbs += 'echo "number of files to be launched: "'
        pbs += '`find "%s" -type f | wc -l`\n' % os.path.join(sim_id, pbs_in_base)
        rpl = (cmd_find, os.path.join(sim_id, pbs_in_base), cmd_xargs, ppn)
        cmd = ("%s '%s' -type f -name '*.p' -print0 | sort -z | %s -0 -I{} "
               "--process-slot-var=CPU_NR -n 1 -P %i sh {}\n" % rpl)
        pbs += cmd
        pbs += 'echo "END OF JOBS IN find+xargs MODE"\n'

        # =====================================================================
        # move results back from the node sim_id dir to the origin
        pbs += '\n'
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'echo "total scratch disk usage:"\n'
        pbs += 'du -hs "%s"\n' % pbase
        pbs += 'cd "%s"\n' % os.path.join(pbase, sim_id)
        pbs += 'echo "current working directory:"\n'
        pbs += 'pwd\n'
        pbs += 'echo "Results saved at sim_id directory:"\n'
        rpl = (os.path.join(pbs_in_base, '*'), os.path.join(htc_base, '*'))
        pbs += 'find .\n'

        if compress:
            # compress all result files into an archive, first *.sel files
            # FIXME: why doesn this work with -name "*.sel" -o -name "*.dat"??
            pbs += '\necho "move results into compressed archive"\n'
            pbs += 'find "%s" -name "*.sel" -print0 ' % res_base
            fname = os.path.join(res_base, 'resfiles_chnk_%05i' % ii)
            pbs += '| xargs -0 tar --remove-files -rf %s.tar\n' % fname
            # now add the *.dat files to the archive
            pbs += 'find "%s" -name "*.dat" -print0 ' % res_base
            fname = os.path.join(res_base, 'resfiles_chnk_%05i' % ii)
            pbs += '| xargs -0 tar --remove-files -rf "%s.tar"\n' % fname
            pbs += 'xz -z2 -T %i "%s.tar"\n' % (ppn, fname)

            # compress all logfiles into an archive
            pbs += '\necho "move logfiles into compressed archive"\n'
            pbs += 'find "%s" -name "*.log" -print0 ' % log_base
            fname = os.path.join(log_base, 'logfiles_chnk_%05i' % ii)
            pbs += '| xargs -0 tar --remove-files -rf "%s.tar"\n' % fname
            pbs += 'xz -z2 -T %i "%s.tar"\n' % (ppn, fname)

        # compress all post-processing results (saved as csv's) into an archive
        pbs += '\necho "move statsdel into compressed archive"\n'
        pbs += 'find "%s" -name "*.csv" -print0 ' % res_base
        fname = os.path.join(post_dir_base, 'statsdel_chnk_%05i' % ii)
        pbs += '| xargs -0 tar --remove-files -rf "%s.tar"\n' % fname
        pbs += 'xz -z2 -T %i "%s.tar"\n' % (ppn, fname)

        # compress all post-processing results (saved as csv's) into an archive
        pbs += '\necho "move log analysis into compressed archive"\n'
        pbs += 'find "%s" -name "*.csv" -print0 ' % log_base
        fname = os.path.join(post_dir_base, 'loganalysis_chnk_%05i' % ii)
        pbs += '| xargs -0 tar --remove-files -rf "%s.tar"\n' % fname
        pbs += 'xz -z2 -T %i "%s.tar"\n' % (ppn, fname)

        pbs += '\n'
        pbs += '\necho "%s"\n' % ('-'*70)
        pbs += 'cd "%s"\n' % pbase
        pbs += "echo 'current working directory:'\n"
        pbs += 'pwd\n'
        pbs += 'echo "move results back from node scratch/sim_id to origin, '
        pbs += 'but ignore htc, and pbs_in directories."\n'

        tmp = os.path.join(sim_id, '') # make we have trailing /
        pbs += 'echo "copy from %s to $PBS_O_WORKDIR/"\n' % tmp
        pbs += 'time rsync -au "%s" "$PBS_O_WORKDIR/" \\\n' % tmp
        pbs += '    --exclude "%s" \\\n' % os.path.join(pbs_in_base, '*')
        pbs += '    --exclude *.htc\n'
        # when using -u, htc and pbs_in files should be ignored
#        pbs += 'time cp -ru %s $PBS_O_WORKDIR/\n' % tmp
        if pyenv is not None:
            pbs += 'source deactivate\n'
        pbs += 'echo "DONE !!"\n'
        pbs += '\necho "%s"\n' % ('-'*70)
        if wine_arch!=None or wine_prefix!=None:
            pbs += '# in case wine has crashed, kill any remaining wine servers\n'
            pbs += '# caution: ALL the users wineservers will die on this node!\n'
            pbs += 'echo "following wineservers are still running:"\n'
            pbs += 'ps -u $USER -U $USER | grep wineserver\n'
            pbs += 'killall -u $USER wineserver\n'
        pbs += 'exit\n'

        rpl = (sim_id, ii)
        fname = os.path.join(run_dir, chunks_dir, '%s_chnk_%05i' % rpl)

        if pbs.find('rm -rf') > -1 or pbs.find('rm -fr') > -1:
            raise UserWarning('Anything that looks like rm -rf is prohibited.')

        with open(fname+'.p', 'w') as f:
            f.write(pbs)

    def make_pbs_postpro_chunks():
        """When only the post-processing has to be re-done for a chunk.
        """
        pass


    cc = Cases(cases)
    df = cc.cases2df()
    # sort on the specified values in the given columns
    # FIXME: sort_values was only added in Pandas 0.17, but CI runner is old
    try:
        df.sort_values(by=sort_by_values, inplace=True)
    except AttributeError:
        df.sort(columns=sort_by_values, inplace=True)

    # create the directory to store all zipped chunks
    try:
        os.mkdir(os.path.join(df['[run_dir]'].iloc[0], chunks_dir))
    # FIXME: how do you make this work pythonically on both PY2 and PY3?
    except (FileExistsError, OSError):
        pass

    fpath = os.path.join(df['[run_dir]'].iloc[0], 'pbs_out_chunks')
    if not os.path.exists(fpath):
        os.makedirs(fpath)

    # remove all cases that start with test_ and move them into a separate
    # chunk
    try:
        sel_notest = df['[Case folder]'].str.find('test_') < 0

        if '[hs2]' in df.columns:
            # depends on if '' or ';' was translated into 0/1 or not
            if df['[hs2]'].dtype == np.dtype('O'):
                sel_hs2 = df['[hs2]'].str.find(';') < 0
            else:
                sel_hs2 = df['[hs2]'] == 1
        else:
            sel_hs2 = pd.Series(data=False, index=df.index)

        # remove hs2 cases from notest
        sel_notest = sel_notest & ~sel_hs2

        # tests but not HS2
        sel_test = ~sel_notest & ~sel_hs2

    except AttributeError:
        # FIXME: CI runner has and old pandas version (v0.14.1)
        # and findall returns list with the search str occuring as
        # many times as found in the str...
        findall = df['[Case folder]'].str.findall('test_').tolist()
        # len==0 if nothing has been found
        sel_notest = [True if len(k)==0 else False for k in findall]

        if '[hs2]' in df.columns:
            if df['[hs2]'].dtype == np.dtype('O'):
                findall = df['[hs2]'].str.findall(';').tolist()
            else:
                findall = (df['[hs2]']==1).tolist()

            # only select it if we do NOT find ;
            sel_hs2 = [True if len(k)==0 else False for k in findall]
        else:
            sel_hs2 = [False]*len(findall)

        # remove hs2 cases from notest
        sel_notest = [i and not j for (i,j) in zip(sel_notest, sel_hs2)]
        # tests but not HS2, so everything that is not in either notest or hs2
        sel_test = [not i and not j for (i,j) in zip(sel_notest, sel_hs2)]

    df_dlc = df[sel_notest]
    df_test = df[sel_test]
    df_hs2 = df[sel_hs2]

    # DLC CHUNKS
    sim_id = df['[sim_id]'].iloc[0]
    run_dir = df['[run_dir]'].iloc[0]
    model_zip = df['[model_zip]'].iloc[0]
    post_dir = df['[post_dir]'].iloc[0]

    names = ['case_id-chunk-index', 'case_id-chunk-test-index',
             'case_id-chunk-hs2-index']
    i0s = [0, 90000, 80000]

    # group test_ hs2 and normal dlc's in 3 different chunks
    for df, name, i02 in zip([df_dlc, df_test, df_hs2], names, i0s):
        i02 = i02 + i0
        df_ind = pd.DataFrame(columns=['chnk_nr'], dtype=np.int32)
        df_ind.index.name = '[case_id]'
        df_iter = chunker(df, nr_procs_series*ppn)
        for ii, dfi in enumerate(df_iter):
            fname, ind = make_zip_chunks(dfi, i02+ii, sim_id, run_dir, model_zip)
            make_pbs_chunks(dfi, i02+ii, sim_id, run_dir, model_zip,
                            wine_arch=wine_arch, wine_prefix=wine_prefix,
                            compress=compress)
            df_ind = pd.concat([df_ind, ind])
            print(fname)
        fname = os.path.join(post_dir, 'case_id-chunk-index')
        df_ind['chnk_nr'] = df_ind['chnk_nr'].astype(np.int32)
        df_ind.to_hdf(fname+'.h5', 'table', complevel=9, complib='zlib')
        df_ind.to_csv(fname+'.csv')


def regroup_tarfiles(cc):
    """Re-group all chunks again per [Case folder] compressed file. First all
    chunks are copied to the node scratch disc, then start working on them.
    This only works on a node with PBS stuff.

    Make sure to maintain the same location as defined by the tags!

    [res_dir] and [Case folder] could be multiple directories deep, bu the
    final archive will only contain the files (no directory structure), and
    the name of the archive is that of the last directory:
        /[res_dir]/[Case folder]/[Case folder].tar.xz
        /res/dir/case/folder/dlcname/dlcname.tar.xz

    Parameters
    ----------

    path_pattern : str
        /path/to/files/*.tar.xz

    """

    USER = os.getenv('USER')
    PBS_JOBID = os.getenv('PBS_JOBID')
    scratch = os.path.join('/scratch', USER, PBS_JOBID)
    src = os.getenv('PBS_O_WORKDIR')

    path_pattern = '/home/dave/SimResults/NREL5MW/D0022/prepost-data/*.xz'

    for ffname in tqdm(glob.glob(path_pattern)):
        appendix = os.path.basename(ffname).split('_')[0]
        with tarfile.open(ffname, mode='r:xz') as tar:
            # create new tar files if necessary for each [Case folder]
            for tarinfo in tar.getmembers():
                t2_name = os.path.basename(os.path.dirname(tarinfo.name))
                t2_dir = os.path.join(os.path.dirname(path_pattern), t2_name)
                if not os.path.isdir(t2_dir):
                    os.makedirs(t2_dir)
                t2_path = os.path.join(t2_dir, t2_name + '_%s.tar' % appendix)
                fileobj = tar.extractfile(tarinfo)
                # change the location of the file in the new archive:
                # the location of the archive is according to the folder
                # structure as defined in the tags, remove any subfolders
                tarinfo.name = os.basename(tarinfo.name)
                with tarfile.open(t2_path, mode='a') as t2:
                    t2.addfile(tarinfo, fileobj)


def merge_from_tarfiles(df_fname, path, pattern, tarmode='r:xz', tqdm=False,
                        header='infer', names=None, sep=',', min_itemsize={},
                        verbose=False, dtypes={}):
    """Merge all csv files from various tar archives into a big pd.DataFrame
    store.

    Parameters
    ----------

    df_fname : str
        file name of the pd.DataFrame h5 store in which all chunks will be
        merged. Names usually used are:
            * [sim_id]_ErrorLogs.h5
            * [sim_id]_statistics.h5

    path : str
        Directory in which all chunks are located.

    pattern : str
        Search pattern used to select (using glob.glob) files in path

    tarmode : str, default='r:xz'
        File opening mode for tarfile (used when opening each of the chunks).

    tqdm : boolean, default=False
       If True, an interactive progress bar will be displayed (requires the
       tqdm module). If set to False no progress bar will be displayed.

    header : int, default='infer'
        Argument passed on to pandas.read_csv. Default to 'infer', set to
        None if there is no header, set to 0 if header is on first row.

    names : list of column names, default=None
        Argument passed on to pandas.read_csv. Default to None. List with
        column names to be used in the DataFrame.

    min_itemsize : dict, default={}
        Argument passed on to pandas.HDFStore.append. Set the minimum lenght
        for a given column in the DataFrame.

    sep : str, default=','
        Argument passed on to pandas.read_csv. Set to ';' when handling the
        ErrorLogs.

    """

    store = pd.HDFStore(os.path.join(path, df_fname), mode='w', format='table',
                        complevel=9, complib='zlib')

    if tqdm:
        from tqdm import tqdm
    else:
        def tqdm(itereable):
            return itereable

    if len(glob.glob(os.path.join(path, pattern))) < 1:
        raise RuntimeError('No files found to merge')

    for tar_fname in tqdm(glob.glob(os.path.join(path, pattern))):
        if verbose:
            print(tar_fname)
        with tarfile.open(tar_fname, mode=tarmode) as tar:
            df = pd.DataFrame()
            for tarinfo in tar.getmembers():
                fileobj = tar.extractfile(tarinfo)
                tmp = pd.read_csv(fileobj, header=header, names=names, sep=sep)
                for col, dtype in dtypes.items():
                    tmp[col] = tmp[col].astype(dtype)
                df = df.append(tmp)
            try:
                if verbose:
                    print('writing...')
                store.append('table', df, min_itemsize=min_itemsize)
            except Exception as e:
                if verbose:
                    print('store columns:')
                    print(store.select('table', start=0, stop=0).columns)
                    print('columns of the DataFrame being added:')
                    print(df.columns)
                storecols = store.select('table', start=0, stop=0).columns
                store.close()
                print(e)
                return df, storecols

    store.close()


# TODO: make this class more general so you can also just give a list of files
# to be merged, excluding the tar archives.
class AppendDataFrames(object):
    """Merge DataFrames, either in h5 or csv format, located in (compressed)
    tar archives.
    """

    def __init__(self, tqdm=False):
        if tqdm:
            from tqdm import tqdm
        else:
            def tqdm(itereable):
                return itereable
        self.tqdm = tqdm

    def _open(self, fname, tarmode='r:xz'):
        """Open text file directly or from a tar archive. Return iterable
        since a tar archive might contain several csv text files
        """

        if fname.find('.tar') > -1:
            with tarfile.open(fname, mode=tarmode) as tar:
                for tarinfo in tar.getmembers():
                    linesb = tar.extractfile(tarinfo).readlines()
                    # convert from bytes to strings
                    lines = [line.decode() for line in linesb]
                    yield lines, tarinfo.name
        else:
            with open(fname, 'r') as f:
                lines = f.readlines()
            yield lines, os.path.basename(fname)

    def df2store(self, store, path, tarmode='r:xz', min_itemsize={},
                 colnames=None, header='infer', columns=None, sep=';',
                 index2col=None, ignore_index=True, fname_col=False):
        """This is very slow, use txt2txt instead.
        """

        # TODO: it seems that with threading you could parallelize this kind
        # of work: http://stackoverflow.com/q/23598063/3156685
        # http://stackoverflow.com/questions/23598063/
        # multithreaded-web-scraper-to-store-values-to-pandas-dataframe

        # http://gouthamanbalaraman.com/blog/distributed-processing-pandas.html

        df = pd.DataFrame()
        for fname in self.tqdm(glob.glob(path)):
            with tarfile.open(fname, mode=tarmode) as tar:
                df = pd.DataFrame()
                for tarinfo in tar.getmembers():
                    fileobj = tar.extractfile(tarinfo)
                    if tarinfo.name[-2:] == 'h5':
                        tmp = pd.read_hdf(fileobj, 'table', columns=columns)
                    elif tarinfo.name[-3:] == 'csv':
                        tmp = pd.read_csv(fileobj, sep=sep, names=colnames,
                                          header=header, usecols=columns)
                    else:
                        continue
                    if index2col is not None:
                        # if the index does not have a name we can still set it
                        tmp[index2col] = tmp.index
                        tmp[index2col] = tmp[index2col].astype(str)
                        tmp.reset_index(level=0, drop=True, inplace=True)
                    # add the file name as a column
                    if fname_col:
                        case_id = os.path.basename(tarinfo.name)
                        tmp[fname_col] = '.'.join(case_id.split('.')[:-1])
                        tmp[fname_col] = tmp[fname_col].astype(str)
                    df = df.append(tmp, ignore_index=ignore_index)

                store.append('table', df, min_itemsize=min_itemsize)
#                if len(df) > w_every:
#                    # and merge into the big ass DataFrame
#                    store.append('table', df, min_itemsize=min_itemsize)
#                    df = pd.DataFrame()
        return store

    # FIXME: when merging log file analysis (files with header), we are still
    # skipping over one case
    def txt2txt(self, fjoined, path, tarmode='r:xz', header=None, sep=';',
                fname_col=False, header_fjoined=None, recursive=False):
        """Read as strings, write to another file as strings.

        Parameters
        ----------

        fjoined

        path

        tarmode

        header : int, default=None
            Indicate if data files contain a header and on which line it is
            located. Set to None if data files do not contain header, and in
            that case the joined file will not contain a header either. All
            lines above the header are ignored.

        sep

        fname_col

        header_fjoined : str, default=None
            If the data files do not contain a header write out header_fjoined
            as the header of the joined file.

        recursive

        Return
        ------

        header_fjoined : str
            String of the header that was written to the joined file.

        """
        if isinstance(header, int):
            write_header = True
            icut = header + 1
        else:
            # when header is None, there is no header
            icut = 0
            write_header = False
        if isinstance(header_fjoined, str):
            write_header = True

        with tempfile.NamedTemporaryFile(mode='w', delete=False) as ft:
            ftname = ft.name
            for fname in self.tqdm(glob.glob(path, recursive=recursive)):
                for lines, case_id in self._open(fname, tarmode=tarmode):
                    # only include the header at the first round
                    if write_header:
                        if header_fjoined is None:
                            header_fjoined = lines[header]
                            header_f0 = lines[header]
                        # add extra column with the file name if applicable
                        if fname_col:
                            rpl = sep + fname_col + '\n'
                            header_fjoined = header_fjoined.replace('\n', rpl)
                        ft.write(header_fjoined)
                        write_header = False
                    # check if the header is the same for each file
                    if header is not None:
                        assert lines[header] == header_f0
                    # but cut out the header on all other occurances
                    case_id = '.'.join(case_id.split('.')[:-1])
                    for line in lines[icut:]:
                        if fname_col:
                            line = line.replace('\n', sep + case_id + '\n')
                        ft.write(line)
                ft.flush()

        # and move from temp dir to fjoined
        shutil.move(ftname, fjoined)
        return header_fjoined.replace('\n', '')

    def csv2df_chunks(self, store, fcsv, chunksize=100000, min_itemsize={},
                      colnames=None, dtypes={}, header='infer', sep=';'):
        """Convert a large csv file to a pandas.DataFrame in chunks using
        a pandas.HDFStore.
        """
        df_iter = pd.read_csv(fcsv, chunksize=chunksize, sep=sep,
                              names=colnames, header=header)
        for df_chunk in self.tqdm(df_iter):
            for col, dtype in dtypes.items():
                df_chunk[col] = df_chunk[col].astype(dtype)
            store.append('table', df_chunk, min_itemsize=min_itemsize)
        return store


if __name__ == '__main__':
    pass



"""Test wetb.hawc2.pc_file
"""
from wetb.hawc2.ae_file import AEFile

import os
import unittest
import tempfile

import numpy as np

from wetb.hawc2.pc_file import PCFile


class TestPCFile(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.testfilepath = os.path.join(os.path.dirname(__file__), 'test_files/')  # test file path

    def test_PCFile_ae(self):
        """Verify correct values in values of object loaded from pc file"""
        pc = PCFile(self.testfilepath + "NREL_5MW_pc.txt")
        ae = AEFile(self.testfilepath + "NREL_5MW_ae.txt")
        thickness = ae.thickness(36)
        self.assertEqual(pc.CL(thickness, 10), 1.358)
        self.assertEqual(pc.CD(thickness, 10), 0.0255)
        self.assertEqual(pc.CM(thickness, 10), -0.1103)

    def test_write_PCFile(self):
        """Round-trip loading and saving a pc file
        """
        pc1 = PCFile(self.testfilepath + 'NREL_5MW_pc.txt')
        with tempfile.TemporaryDirectory() as tdir:
            pc1.save(tdir + '/test_pc.txt')
            pc2 = PCFile(pc1.filename)
            self.assertEqual(str(pc1), str(pc2))

        tc1, pcs1 = pc1.pc_sets[1]
        tc2, pcs2 = pc2.pc_sets[1]
        np.testing.assert_array_almost_equal(tc1, tc2)
        for pc1, pc2 in zip(pcs1, pcs2):
            np.testing.assert_array_almost_equal(pc1, pc2)



if __name__ == "__main__":
    unittest.main()
